"""
    Copyright 2009-2023 Stephane De Mita, Mathieu Siol, Thomas Coudoux

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

# cf: https://www.ncbi.nlm.nih.gov/books/NBK279684/
# see also: https://doctorlib.info/medical/blast/7.html (checked for strands)

import subprocess, os, collections, sys
import xml.etree.ElementTree
from .. import _interface, alphabets
from ..tools import _code_tools
from . import _utils

class _BLAST_tool(_utils._App):
    def _check_path(self, path, cfg):
        cmd = (path, '-version')
        try:
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
            stdout, stderr = p.communicate()
            if len(stderr): return stderr
        except OSError as e:
            return e.strerror

_blast_apps = {}
for k in 'makeblastdb', 'blastn', 'blastp', 'blastx', 'tblastn', 'tblastx':
    _blast_apps[k] = _BLAST_tool(key=k, default=k)
    _utils.paths._add(_blast_apps[k])

def makeblastdb(source, dbtype=None, out=None, input_type="fasta", verbose=False,
    title=None, parse_seqids=False, hash_index=False, mask_data=None, mask_id=None,
    mask_desc=None, blastdb_version=5,
    max_file_sz="1GB", taxid=None, taxid_map=None):
    """
    Create a BLAST database.

    :param source: name of an input file of the appropriate format.
        If not fasta, the format must be specified using the *input_type*
        option. Alternatively, *source* can be a :class:`.Container` or
        :class:`.Align` instance. If so, its alphabet must be DNA or protein
        and the *dbtype* argument, if specified, must match. Note that
        passing a :class:`!Container` or :class:`!Align` instance must be
        avoided for large databases.

    :param dbtype: database type: ``"nucl"`` or ``"prot"`` are acceptable. Can
        be omitted if a :class:`!Container` or an :class:`!Align` is provided
        as *source*.

    :param out: database name. Must be specified if a :class:`!Container`
        or an :class:`!Align` is provided as *source*, or if the *input_type*
        is "blastdb", otherwise the input file name is used as database name.

    :param input_type: format of input file. Must be ``"fasta"`` if
        a :class:`!Container` or an :class:`!Align` is provided as *source*.
        Otherwise must describe the format of *source*: ``"fasta"``, ``"asn1_bin"``,
        ``"asn1_txt"``, or ``"blastdb"``.

    :param verbose: display makeblastdb output (by default, it is returned
        by the function). Errors are always displayed.

    :param title: database title. A default title is inserted in case a
        :class:`!Container` or :class:`!Align` instance is passed as *source*.

    :parse_seqids: parse seqid from sequence names (considered if `input_type` is fasta,
        including if a :class:`!Container` or an :class:`!Align` is provided
        as *source*; argument ignored otherwise: seqid is always imported).

    :hash_index: create index of sequence hash values.

    :mask_data: list of input files containing masking data.

    :mask_id: list of strings to uniquely identify the masking algorithm,
              one for each mask file (requires *mask_data*).

    :mask_desc: list of free form strings to describe the masking algorithm
                details, one for each mask file (requires *mask_id*).

    :blastdb_version: version of BLAST database to be created (4 or 5).

    :max_file_sz: maximum file size for BLAST database files.

    :taxid: taxonomy ID to assign to all sequences as an integer
            (incompatible with *taxid_map*).

    :taxid_map: text file mapping sequence IDs to taxonomy IDs
            (requires *parse_seqids*, incompatible with *taxid*).

    :return: Standard output of the program (`None` if *verbose* was ``True``).

    Please refer to the manual of BLAST tools for more details.

    .. versionchanged 3.2::
        *blastdb_version* default value was previously 4. *gi_mask* and
        *gi_mask_name* options are removed.
    """

    # get command name
    path = _blast_apps['makeblastdb'].get_path()
    if path is None:
        raise RuntimeError('makeblastdb program not available -- please configure path')
    cmd = [path]

    # process main arguments arguments
    if isinstance(source, (_interface.Container, _interface.Align)):
        cmd.extend(['-in', '-'])
        if source.alphabet == alphabets.DNA and (dbtype is None or dbtype == 'nucl'): cmd.extend(['-dbtype', 'nucl'])
        elif source.alphabet == alphabets.protein and (dbtype is None or dbtype == 'prot'): cmd.extend(['-dbtype', 'prot'])
        else: raise ValueError('invalid alphabet, dbtype or alphabet/dbtype mismatch')
        if input_type != 'fasta': raise ValueError('`input_type` must be "fasta"')
        if out is None: raise ValueError('`out` is required if an object is passed as `source`')
        cmd.extend(['-input_type', 'fasta'])
        if title is None: title = '{0} database from an EggLib {1}'.format(cmd[4], 'Container' if isinstance(source, _interface.Container) else 'Align')
    else:
        if not isinstance(source, (os.PathLike, str)): raise TypeError('`source` must be a file name, a Container or an Align')
        if input_type != 'blastdb' and not os.path.isfile(source): raise ValueError('file not found: {0}'.format(source))
        if dbtype is None: raise ValueError('`dbtype` is required')
        if dbtype not in ['nucl', 'prot']: raise ValueError('invalid value for `dbtype`')
        if input_type not in ['fasta', 'asn1_bin', 'asn1_txt', 'blastdb']: raise ValueError('invalid value for `input_type`')
        if input_type == 'blastdb' and out is None or out == source: raise ValueError('a different database name is required for input type `blastdb`')
        cmd.extend(['-in', source, '-dbtype', dbtype, '-input_type', input_type])
    if out is not None:
        if not isinstance(out, (os.PathLike, str)): raise TypeError('`out` must be a string or a path-like object')
        cmd.extend(['-out', out])

    # check types of other arguments
    if not isinstance(blastdb_version, int): raise TypeError('`blastdb_version` must be an integer')
    if blastdb_version not in [4, 5]: raise ValueError('supported values for `blastdb_version` are 4 and 5 only')
    if not isinstance(max_file_sz, str): raise TypeError('`max_file_sz` must be a string')
    if title is not None and not isinstance(title, str): raise TypeError('`title` must be a string')
    if not isinstance(parse_seqids, bool): raise TypeError('`parse_seqids` must be a boolean')
    if input_type != 'fasta': parse_seqids = True
    if not isinstance(hash_index, bool): raise TypeError('`hash_index` must be a boolean')
    if mask_data is not None:
        if not isinstance(mask_data, (list, tuple)): raise TypeError('`mask_data` must be a list or a tuple')
        if not all([isinstance(item, str) for item in mask_data]): raise TypeError('`mask_data` items must be strings')
        if not all(map(os.path.isfile, mask_data)): raise ValueError('at least one file specified in `mask_data` does not exist')
        if len(mask_data) == 0: raise ValueError('there must be at least one item in `mask_data`')
    if mask_id is not None:
        if not isinstance(mask_id, (list, tuple)): raise TypeError('`mask_id` must be a list or a tuple')
        if not all([isinstance(item, str) for item in mask_id]): raise TypeError('`mask_id` items must be strings')
    if mask_desc is not None:
        if not isinstance(mask_desc, (list, tuple)): raise TypeError('`mask_desc` must be a list or a tuple')
        if not all([isinstance(item, str) for item in mask_desc]): raise TypeError('`mask_desc` items must be strings')
    if taxid is not None:
        if not isinstance(taxid, int): raise TypeError('`taxid` must be an integer')
        if taxid < 0: raise ValueError('`taxid` must be >= 0')
    if taxid_map is not None:
        if not isinstance(taxid_map, (str, os.PathLike)): raise TypeError('`taxid_map` must be a string or pathlike object')
        if not os.path.isfile(taxid_map): raise ValueError('file passed for `taxid_map` not found: {0}'.format(taxid_map))

    # check compatibility of other arguments
    if mask_id is not None:
        if mask_data is None: raise ValueError('`mask_id` requires `mask_data`')
        if len(mask_id) != len(mask_data): raise ValueError('`mask_id` must have the same length than `mask_data`')
    if mask_desc is not None:
        if mask_id is None: raise ValueError('`mask_desc` requires `mask_id`')
        if len(mask_desc) != len(mask_data): raise ValueError('`mask_desc` must have the same length than `mask_data`')
    if taxid is not None and taxid_map is not None: raise ValueError('`taxid` and `taxid_map` are incompatible')
    if taxid_map is not None and not parse_seqids: raise ValueError('`taxid_map` requires `parse_seqids`')

    # load the options
    if title is not None: cmd.extend(['-title', title])
    if parse_seqids: cmd.append('-parse_seqids')
    if hash_index: cmd.append('-hash_index')
    if mask_data is not None: cmd.extend(['-mask_data', ','.join(mask_data)])
    if mask_id is not None: cmd.extend(['-mask_id', ','.join(mask_id)])
    if mask_desc is not None: cmd.extend(['-mask_desc', ','.join(mask_desc)])
    cmd.extend(['-blastdb_version', str(blastdb_version), '-max_file_sz', max_file_sz])
    if taxid: cmd.extend(['-taxid', str(taxid)])
    if taxid_map: cmd.extend(['-taxid_map', taxid_map])

    # run the program
    p = subprocess.Popen(cmd, stdin=subprocess.PIPE,
            stdout=None if verbose else subprocess.PIPE,
            stderr=subprocess.PIPE, universal_newlines=True)
    if isinstance(source, (_interface.Container, _interface.Align)): stdin = source.fasta()
    else: stdin = None
    stdout, stderr = p.communicate(stdin)
    if p.returncode != 0:
        raise RuntimeError('error while running makeblastdb (message below)\n' + stderr)
    return stdout

_blastn_costs = {
        # r/p     default, allowed
        (1,-2): ( (5,2),   [(5,2), (2,2), (1,2), (0,2), (3,1), (2,1), (1,1)] ),
        (1,-3): ( (5,2),   [(5,2), (2,2), (1,2), (0,2), (2,1), (1,1)] ),
        (1,-4): ( (5,2),   [(5,2), (1,2), (0,2), (2,1), (1,1)] ),
        (2,-3): ( (5,2),   [(4,4), (2,4), (0,4), (3,3), (6,2), (5,2), (4,2), (2,2)] ),
        (4,-5): ( (12,8),  [(12,8), (6,5), (5,5), (4,5), (3,5)] ),
        (1,-1): ( (5,2),   [(5,2), (3,2), (2,2), (1,2), (0,2), (4,1), (3,1), (2,1)] )}

_blastp_costs = {
    'BLOSUM62': [(32767, 32767), (11, 2), (10, 2), (9, 2), (8, 2), (7, 2),
                        (6, 2), (13, 1), (12, 1), (11, 1), (10, 1), (9, 1)],

    'BLOSUM80': [(32767, 32767), (25, 2), (13, 2), (9, 2), (8, 2), (7, 2),
                        (6, 2), (11, 1), (10, 1), (9, 1)],

    'PAM30': [(32767, 32767), (7, 2), (6, 2), (5, 2), (10, 1), (9, 1),
                        (8, 1), (15, 3), (14, 2), (14, 1), (13, 3)],

    'PAM70': [(32767, 32767), (8, 2), (7, 2), (6, 2), (11, 1), (10, 1),
                        (9, 1), (11, 2), (12, 3)]
}

_blastp_costs_defaults = {
    'BLOSUM62': (11, 1),
    'BLOSUM80': (10, 1),
    'PAM70': (10, 1),
    'PAM30': (9, 1)
}

def megablast(query, db=None, subject=None, query_loc=None,
        subject_loc=None, evalue=10, parse_deflines=False,
        num_threads=1, word_size=28, gapopen=5, gapextend=2, reward=1,
        penalty=-2, strand='both', no_dust=False, no_soft_masking=False,
        lcase_masking=False, perc_identity=0, no_greedy=False):
    """
    ``megablast`` similarity search. This is designed for strongly similar
    sequences using a nucleotide query on a nucleotide database.

    .. include:: blast_options_common.rst
    .. include:: blast_options_blastn.rst
    :return: A :class:`.BlastOutput` instance.

    """
    return _blastn(query, db, subject, 'megablast', query_loc,
                subject_loc, evalue, parse_deflines, num_threads,
                word_size, gapopen, gapextend, reward, penalty, strand,
                no_dust, no_soft_masking, lcase_masking, perc_identity,
                None, None, no_greedy)

def dc_megablast(query, db=None, subject=None, query_loc=None,
        subject_loc=None, evalue=10, parse_deflines=False,
        num_threads=1, word_size=11, gapopen=None, gapextend=None, reward=2,
        penalty=-3, strand='both', no_dust=False, no_soft_masking=False,
        lcase_masking=False, perc_identity=0, template_type='coding',
        template_length=18):
    """
    Dicontinuous ``megablast`` similarity search. This is designed for similar
    sequences (less similar than :func:`.megablast`) using a nucleotide
    query on a nucleotide database.

    .. include:: blast_options_common.rst
    .. include:: blast_options_blastn.rst
    :param template_type: template type for for dc-megablast. Possible
        values are ``"coding"`` (default), ``"optimal"``, and ``"coding_and_optimal"``.
    :param template_length: template length for dc-megablast. Possible
        values are 16, 18 (the default), and 21.
    """
    return _blastn(query, db, subject, 'dc-megablast', query_loc,
               subject_loc, evalue,parse_deflines, num_threads,
               word_size, gapopen, gapextend, reward, penalty, strand,
               no_dust, no_soft_masking, lcase_masking, perc_identity,
               template_type, template_length, False)

def blastn(query, db=None, subject=None, query_loc=None,
        subject_loc=None, evalue=10, parse_deflines=False,
        num_threads=1, word_size=11, gapopen=None, gapextend=None, reward=2,
        penalty=-3, strand='both', no_dust=False, no_soft_masking=False,
        lcase_masking=False, perc_identity=0):
    """
    ``blastn`` similarity search. This is designed for distant sequences using a
    nucleotide query on a nucleotide database.

    .. include:: blast_options_common.rst
    .. include:: blast_options_blastn.rst
    :return: A :class:`BlastOutput` instance.
    """
    return _blastn(query, db, subject, 'blastn', query_loc, subject_loc,
               evalue, parse_deflines, num_threads, word_size, gapopen,
               gapextend, reward, penalty, strand, no_dust,
               no_soft_masking, lcase_masking, perc_identity, None,
               None, False)

def blastn_short(query, db=None, subject=None, query_loc=None,
        subject_loc=None, evalue=1000, parse_deflines=False,
        num_threads=1, word_size=7, gapopen=None, gapextend=None, reward=1,
        penalty=-3, strand='both', no_dust=True, no_soft_masking=True,
        lcase_masking=False, perc_identity=0):
    """
    ``blastn`` for short sequences. This is optimised for query sequences
    up to 50 bp long. It automatically sets ``evalue=1000``, ``word_size=7``,
    ``no_dust=True``, ``no_soft_masking=True``, ``reward=1``, ``penalty=-3``,
    ``gapopen=5`` and ``gapextend=2``.

    .. include:: blast_options_common.rst
    .. include:: blast_options_blastn.rst
    :return: A :class:`BlastOutput` instance.
    """
    return _blastn(query, db, subject, 'blastn-short', query_loc,
               subject_loc, evalue, parse_deflines, num_threads,
               word_size, gapopen, gapextend, reward, penalty, strand,
               no_dust, no_soft_masking, lcase_masking, perc_identity,
               None, None, False)

@_utils._protect_run
def _blastn(query, db, subject, task, query_loc, subject_loc, evalue,
    parse_deflines, num_threads, word_size, gapopen, gapextend, reward,
    penalty, strand, no_dust, no_soft_masking, lcase_masking,
    perc_identity, template_type, template_length, no_greedy):

    cmd, stdin = _get_common_blast_options('blastn', query, db, subject,
                        query_loc, subject_loc, evalue, parse_deflines, num_threads)

    # process task
    cmd.extend(['-task', task])

    # word size
    if not isinstance(word_size, int): raise TypeError('`word_size` must be an integer')
    if task == 'megablast' and word_size < 4: raise ValueError('invalid value for `word_size`: must be >= 4 for megablast')
    if task == 'dc-megablast' and word_size not in [11, 12]: raise ValueError('invalid value for `word_size`: must be 11 or 12 for dc-megablast')
    if task == 'blastn' and word_size < 4: raise ValueError('invalid value for `word_size`: must be >= 4 for blastn')
    if task == 'blastn-short' and word_size < 4: raise ValueError('invalid value for `word_size`: must be >= 4 for blastn-short')
    cmd.extend(['-word_size', str(word_size)])

    # reward/penalty
    if not isinstance(reward, int): raise TypeError('`reward` must be an integer')
    if reward < 0: raise ValueError('`reward` must be >= 0')
    if not isinstance(penalty, int): raise TypeError('`penalty` must be an integer')
    if penalty > 0: raise ValueError('`penalty` must be <= 0')
    if (reward, penalty) not in _blastn_costs:
        raise ValueError('invalid set of values for reward/penalty: ({0},{1})'.format(reward, penalty))

    # gap costs
    if gapopen is None:
        if gapextend is not None: raise ValueError('if `gapopen` is ``None``, `gapextend` should be also')
        gapopen, gapextend = _blastn_costs[(reward, penalty)][0]
    elif gapextend is None:
        raise ValueError('if `gapextend` is ``None``, `gapopen` should be also')

    if not isinstance(gapopen, int): raise TypeError('`gapopen` must be an integer')
    if gapopen < 0: raise ValueError('`gapopen` must be >= 0')
    if not isinstance(gapextend, int): raise TypeError('`gapextend` must be an integer')
    if gapextend < 0: raise ValueError('`gapextend` must be >= 0')
    if (gapopen, gapextend) not in _blastn_costs[(reward, penalty)][1]:
        raise ValueError('invalid set of values for gapopen/gapextend: ({0},{1})'.format(gapopen, gapextend))
    cmd.extend(['-gapopen', str(gapopen), '-gapextend', str(gapextend),
                '-reward', str(reward), '-penalty', str(penalty)])

    # strand
    if not isinstance(strand, str): raise TypeError('`strand` must be a string')
    if strand not in ['both', 'minus', 'plus']: raise ValueError('`strand` must be either "both", "minus", or "plus"')
    cmd.extend(['-strand', strand])

    # filter/mask flags
    if no_dust: cmd.extend(['-dust', 'no'])
    if no_soft_masking: cmd.extend(['-soft_masking', 'false'])
    if lcase_masking: cmd.append('-lcase_masking')

    #  perc_identity
    if not isinstance(perc_identity, int): raise TypeError('`perc_identity` must be an integer')
    if perc_identity < 0 or perc_identity > 100: raise ValueError('`perc_identity` out of range')
    cmd.extend(['-perc_identity', str(perc_identity)])

    # template parameters
    if task == 'dc-megablast':
        if not isinstance(template_type, str): raise TypeError('`template_type` must be a string')
        if template_type not in ['coding', 'optimal', 'coding_and_optimal']: raise ValueError('invalid value for `template_type`')
        if not isinstance(template_length, int): raise TypeError('`template_length` must be an integer')
        if template_length not in [16, 18, 21]: raise ValueError('invalid value for `template_length`')
        cmd.extend(['-template_type', template_type])
        cmd.extend(['-template_length', str(template_length)])

    # aligment flags
    if no_greedy: cmd.append('-no_greedy')

    # run the program
    p = subprocess.Popen(cmd, stdin=None if stdin is None else subprocess.PIPE,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    stdout, stderr = p.communicate(stdin)
    if p.returncode != 0:
        raise RuntimeError('error while running blastn (message below)\n' + stderr)
    return _parse_blast_output(stdout)

def blastp(query, db=None, subject=None, query_loc=None, subject_loc=None,
        evalue=10, parse_deflines=False, num_threads=1, word_size=None,
        gapopen=None, gapextend=None, matrix='BLOSUM62', threshold=11,
        comp_based_stats=2, seg=0, soft_masking=False,
        lcase_masking=False, window_size=40, use_sw_tback=False):
    """
    ``bastp`` similary search. This is designed for using a protein query
     on a protein database.

    .. include:: blast_options_common.rst
    .. include:: blast_options_prot.rst

    :param parse_deflines: parse query and subject bar delimited
        sequence identifiers.
    :param comp_based_stats: composition-based statistics, as an integer
        code: 0 (no composition-based statistics), 1 (composition-based
        statistics as in NAR 29:2994-3005, 2001), 2 (composition-based
        score adjustment as in Bioinformatics 21:902-911, 2005,
        conditioned on sequence properties), or 3 (composition-based
        score adjustment as in Bioinformatics 21:902-911, 2005,
        unconditionally).
    :param soft_masking: apply filtering locations as soft masks.
    :param lcase_masking: use lower case filtering in query and subject
        sequences.
    :param use_sw_tback: compute locally optimal Smith-Waterman
        alignments.

    :return: A :class:`BlastOutput` instance.
    """
    return _blastp(query, db, subject, 'blastp', query_loc, subject_loc,
        evalue, parse_deflines, num_threads, word_size, gapopen,
        gapextend, matrix, threshold, comp_based_stats, seg,
        soft_masking, lcase_masking, window_size, use_sw_tback)

def blastp_short(query, db=None, subject=None, query_loc=None,
        subject_loc=None, evalue=10, parse_deflines=False,
        num_threads=1, word_size=None, gapopen=None, gapextend=None,
        matrix='PAM30', threshold=16, comp_based_stats=0, seg=0,
        lcase_masking=False, window_size=15, use_sw_tback=False):
    """
    ``blastp`` similarity search for short sequences.

    .. include:: blast_options_common.rst
    .. include:: blast_options_prot.rst

    :param parse_deflines: parse query and subject bar delimited
        sequence identifiers.
    :param comp_based_stats: composition-based statistics, as an integer
        code: 0 (no composition-based statistics), 1 (composition-based
        statistics as in NAR 29:2994-3005, 2001), 2 (composition-based
        score adjustment as in Bioinformatics 21:902-911, 2005,
        conditioned on sequence properties), or 3 (composition-based
        score adjustment as in Bioinformatics 21:902-911, 2005,
        unconditionally).
    :param lcase_masking: use lower case filtering in query and subject
        sequences.
    :param use_sw_tback: compute locally optimal Smith-Waterman
        alignments.

    :return: A :class:`BlastOutput` instance.
    """
    return _blastp(query, db, subject, 'blastp-short', query_loc, subject_loc,
        evalue, parse_deflines, num_threads, word_size, gapopen,
        gapextend, matrix, threshold, comp_based_stats, seg, False,
        lcase_masking, window_size, use_sw_tback)

def blastp_fast(query, db=None, subject=None, query_loc=None,
        subject_loc=None, evalue=10, parse_deflines=False,
        num_threads=1, word_size=None, threshold=21, comp_based_stats=2,
        seg=0, lcase_masking=False, window_size=40, use_sw_tback=False):
    """
    Quick ``blastp`` similarity search.

    .. include:: blast_options_common.rst

    :param threshold: minimum word score such that the word is added to
        the BLAST lookup table (>0).
    :param seg: filter query sequence with SEG as an integer (0 to
        disable, 1 to enable, or alternatively a tuple with the three
        parameters *window*, *locut*, and *hicut*).
    :param window_size: multiple hits window size (use 0 to specify
        1-hit algorithm).
    :param parse_deflines: parse query and subject bar delimited
        sequence identifiers.
    :param comp_based_stats: composition-based statistics, as an integer
        code: 0 (no composition-based statistics), 1 (composition-based
        statistics as in NAR 29:2994-3005, 2001), 2 (composition-based
        score adjustment as in Bioinformatics 21:902-911, 2005,
        conditioned on sequence properties), or 3 (composition-based
        score adjustment as in Bioinformatics 21:902-911, 2005,
        unconditionally).
    :param lcase_masking: use lower case filtering in query and subject
        sequences.
    :param use_sw_tback: compute locally optimal Smith-Waterman
        alignments.

    :return: A :class:`BlastOutput` instance.
    """
    return _blastp(query, db, subject, 'blastp-fast', query_loc, subject_loc,
        evalue, parse_deflines, num_threads, word_size, None,
        None, None, threshold, comp_based_stats, seg, False,
        lcase_masking, window_size, use_sw_tback)

@_utils._protect_run
def _blastp(query, db, subject, task, query_loc, subject_loc, evalue,
    parse_deflines, num_threads, word_size, gapopen, gapextend, matrix,
    threshold, comp_based_stats, seg, soft_masking, lcase_masking,
    window_size, use_sw_tback):

    cmd, stdin = _get_common_blast_options('blastp', query, db, subject,
                        query_loc, subject_loc, evalue, parse_deflines, num_threads)
    _get_common_protein_options(cmd, task, word_size, matrix, threshold,
        gapopen, gapextend, seg, window_size, comp_based_stats)

    # process task
    cmd.extend(['-task', task])

    # filter/mask flags
    if task == 'blastp':
        if soft_masking: cmd.extend(['-soft_masking', 'true'])
        else: cmd.extend(['-soft_masking', 'false'])
    if lcase_masking: cmd.append('-lcase_masking')

    # use_sw_tback
    if use_sw_tback:
        cmd.append('-use_sw_tback')

    # run the program
    p = subprocess.Popen(cmd, stdin=None if stdin is None else subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    stdout, stderr = p.communicate(stdin)
    if  p.returncode != 0:
        raise RuntimeError('error while running blastp (message below)\n' + stderr)
    return _parse_blast_output(stdout)

def blastx(query, db=None, subject=None, query_loc=None, subject_loc=None,
        evalue=10, num_threads=1, word_size=None, gapopen=None,
        gapextend=None, matrix='BLOSUM62', threshold=12,
        seg=(12, 2.2, 2.5), soft_masking=False, lcase_masking=False,
        window_size=40, strand='both', query_genetic_code=1,
        max_intron_length=0, comp_based_stats=2):
    """
    ``blastx`` similarity search. Designed for using a translated nucleotide
    query on a protein database.

    .. include:: blast_options_common.rst
    .. include:: blast_options_prot.rst
    :param soft_masking: apply filtering locations as soft masks.
    :param lcase_masking: use lower case filtering in query and subject
        sequences.
    :param query_genetic_code: genetic code to translate query. Allowed
        values are: 1-6, 9-16, 21-25.
    :param max_intron_length: length of the largest intron allowed in a
        translated nucleotide sequence when linking multiple distinct
        alignments (a negative value disables linking).
    :param strand: query strand(s) to search against database/subject.
        Choice of both, minus, or plus.
    :param comp_based_stats: composition-based statistics, as an integer
        code: 0 (no composition-based statistics), 1 (composition-based
        statistics as in NAR 29:2994-3005, 2001), 2 (composition-based
        score adjustment as in Bioinformatics 21:902-911, 2005,
        conditioned on sequence properties), or 3 (composition-based
        score adjustment as in Bioinformatics 21:902-911, 2005,
        unconditionally).

    :return: A :class:`BlastOutput` instance.

"""
    return _blastx('blastx', query, db, subject, query_loc, subject_loc,
        evalue, False, num_threads, word_size, gapopen,
        gapextend, matrix, threshold, seg, soft_masking, lcase_masking,
        window_size, strand, query_genetic_code, max_intron_length,
        comp_based_stats)

def blastx_fast(query, db=None, subject=None, query_loc=None,
        subject_loc=None, evalue=10,  num_threads=1, word_size=None,
        gapopen=None, gapextend=None, matrix='BLOSUM62', threshold=21,
        seg=(12, 2.2, 2.5), soft_masking=False, lcase_masking=False,
        window_size=40, strand='both', query_genetic_code=1,
        max_intron_length=0, comp_based_stats=2):
    """
    Quick ``blastx`` similarity search. Designed for using a translated nucleotide
    query on a protein database and optimised for faster execution.

    .. include:: blast_options_common.rst
    .. include:: blast_options_prot.rst
    :param soft_masking: apply filtering locations as soft masks.
    :param lcase_masking: use lower case filtering in query and subject
        sequences.
    :param query_genetic_code: genetic code to translate query. Allowed
        values are: 1-6, 9-16, 21-25.
    :param max_intron_length: length of the largest intron allowed in a
        translated nucleotide sequence when linking multiple distinct
        alignments (a negative value disables linking).
    :param strand: query strand(s) to search against database/subject.
        Choice of both, minus, or plus.
    :param comp_based_stats: composition-based statistics, as an integer
        code: 0 (no composition-based statistics), 1 (composition-based
        statistics as in NAR 29:2994-3005, 2001), 2 (composition-based
        score adjustment as in Bioinformatics 21:902-911, 2005,
        conditioned on sequence properties), or 3 (composition-based
        score adjustment as in Bioinformatics 21:902-911, 2005,
        unconditionally).

    :return: A :class:`BlastOutput` instance.
    """
    return _blastx('blastx-fast', query, db, subject, query_loc,
        subject_loc, evalue, False, num_threads, word_size,
        gapopen, gapextend, matrix, threshold, seg, soft_masking,
        lcase_masking, window_size, strand, query_genetic_code,
        max_intron_length, comp_based_stats)

@_utils._protect_run
def _blastx(task, query, db, subject, query_loc,
        subject_loc, evalue, parse_deflines, num_threads, word_size,
        gapopen, gapextend, matrix, threshold, seg, soft_masking,
        lcase_masking, window_size, strand, query_genetic_code,
        max_intron_length, comp_based_stats):

    cmd, stdin = _get_common_blast_options('blastx', query, db, subject,
        query_loc, subject_loc, evalue, parse_deflines, num_threads)
    cmd.extend(['-task', task])

    _get_common_protein_options(cmd, task, word_size, matrix, threshold,
        gapopen, gapextend, seg, window_size, comp_based_stats)

    if soft_masking: cmd.extend(['-soft_masking', 'true'])
    else: cmd.extend(['-soft_masking', 'false'])
    if lcase_masking: cmd.append('-lcase_masking')

    if not isinstance(strand, str): raise TypeError('`strand` must be a string')
    if strand not in ['both', 'minus', 'plus']: raise ValueError('`strand` must be either "both", "minus", or "plus"')
    cmd.extend(['-strand', strand])

    if query_genetic_code not in [1, 2, 3, 4, 5, 6, 9, 10, 11, 12, 13, 14, 15, 16, 21, 22, 23, 24, 25]:
        raise ValueError('invalid genetic code: {0}'.format(query_genetic_code))
    cmd.extend(['-query_gencode', str(query_genetic_code)])

    if not isinstance(max_intron_length, int): raise TypeError('`max_intron_length` must be an integer')
    if max_intron_length < 0: raise ValueError('`max_intron_length` must be >= 0')
    cmd.extend(['-max_intron_length', str(max_intron_length)])

    # run the program
    p = subprocess.Popen(cmd, stdin=None if stdin is None else subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    stdout, stderr = p.communicate(stdin)
    if p.returncode != 0:
        raise RuntimeError('error while running blastx (message below)\n' + stderr)
    return _parse_blast_output(stdout)

def tblastn(query, db=None, subject=None, query_loc=None, subject_loc=None,
        evalue=10, num_threads=1, word_size=None, gapopen=None,
        gapextend=None, matrix='BLOSUM62', threshold=13,
        seg=(12, 2.2, 2.5), soft_masking=False, window_size=40,
        db_genetic_code=1, max_intron_length=0, comp_based_stats=2):
    """
    ``tblastn`` similary search. Designed for using a protein query on
    a translated nucleotide database.

    .. include:: blast_options_common.rst
    .. include:: blast_options_prot.rst

    :param soft_masking: apply filtering locations as soft masks.
    :param db_genetic_code: genetic code to translate subject sequences.
        Allowed values are: 1-6, 9-16, 21-25.
    :param max_intron_length: length of the largest intron allowed in a
        translated nucleotide sequence when linking multiple distinct
        alignments (a negative value disables linking).
    :param comp_based_stats: composition-based statistics, as an integer
        code: 0 (no composition-based statistics), 1 (composition-based
        statistics as in NAR 29:2994-3005, 2001), 2 (composition-based
        score adjustment as in Bioinformatics 21:902-911, 2005,
        conditioned on sequence properties), or 3 (composition-based
        score adjustment as in Bioinformatics 21:902-911, 2005,
        unconditionally).

    :return: A :class:`BlastOutput` instance.
    """
    return _tblastn('tblastn', query, db, subject, query_loc, subject_loc,
        evalue, num_threads, word_size, gapopen, gapextend, matrix,
        threshold, seg, soft_masking, window_size, db_genetic_code,
        max_intron_length, comp_based_stats)

def tblastn_fast(query, db=None, subject=None, query_loc=None, subject_loc=None,
        evalue=10, num_threads=1, word_size=None, gapopen=None,
        gapextend=None, matrix='BLOSUM62', threshold=21,
        seg=(12, 2.2, 2.5), soft_masking=False, window_size=40,
        db_genetic_code=1, max_intron_length=0, comp_based_stats=2):
    """
    Quich ``tblastn`` similary search. Designed for using a protein query on
    a translated nucleotide database and optimised for fast execution.

    .. include:: blast_options_common.rst
    .. include:: blast_options_prot.rst

    :param soft_masking: apply filtering locations as soft masks.
    :param db_genetic_code: genetic code to translate subject sequences.
        Allowed values are: 1-6, 9-16, 21-25.
    :param max_intron_length: length of the largest intron allowed in a
        translated nucleotide sequence when linking multiple distinct
        alignments (a negative value disables linking).
    :param comp_based_stats: composition-based statistics, as an integer
        code: 0 (no composition-based statistics), 1 (composition-based
        statistics as in NAR 29:2994-3005, 2001), 2 (composition-based
        score adjustment as in Bioinformatics 21:902-911, 2005,
        conditioned on sequence properties), or 3 (composition-based
        score adjustment as in Bioinformatics 21:902-911, 2005,
        unconditionally).

    :return: A :class:`BlastOutput` instance.
    """
    return _tblastn('tblastn-fast', query, db, subject, query_loc, subject_loc,
        evalue, num_threads, word_size, gapopen, gapextend, matrix,
        threshold, seg, soft_masking, window_size, db_genetic_code,
        max_intron_length, comp_based_stats)

@_utils._protect_run
def _tblastn(task, query, db, subject, query_loc, subject_loc,
        evalue, num_threads, word_size, gapopen, gapextend, matrix,
        threshold, seg, soft_masking, window_size, db_genetic_code,
        max_intron_length, comp_based_stats):

    # process common options
    cmd, stdin = _get_common_blast_options('tblastn', query, db, subject,
                        query_loc, subject_loc, evalue, False, num_threads)
    _get_common_protein_options(cmd, task, word_size, matrix, threshold,
        gapopen, gapextend, seg, window_size, comp_based_stats)

    # process task
    cmd.extend(['-task', task])

    # filter/mask flags
    if soft_masking: cmd.extend(['-soft_masking', 'true'])
    else: cmd.extend(['-soft_masking', 'false'])

    # db_genetic_code
    if db_genetic_code not in [1, 2, 3, 4, 5, 6, 9, 10, 11, 12, 13, 14, 15, 16, 21, 22, 23, 24, 25]:
        raise ValueError('invalid genetic code: {0}'.format(db_genetic_code))
    cmd.extend(['-db_gencode', str(db_genetic_code)])

    # max_intron_length
    if not isinstance(max_intron_length, int): raise TypeError('`max_intron_length` must be an integer')
    if max_intron_length < 0: raise ValueError('`max_intron_length` must be >= 0')
    cmd.extend(['-max_intron_length', str(max_intron_length)])

    # run the program
    p = subprocess.Popen(cmd, stdin=None if stdin is None else subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    stdout, stderr = p.communicate(stdin)
    if p.returncode != 0:
        raise RuntimeError('error while running tblastn (message below)\n' + stderr)
    return _parse_blast_output(stdout)

@_utils._protect_run
def tblastx(query, db=None, subject=None, query_loc=None, subject_loc=None,
        evalue=10, num_threads=1, word_size=None, matrix='BLOSUM62',
        seg=(12, 2.2, 2.5), soft_masking=False, lcase_masking=False,
        window_size=40, db_genetic_code=1, query_genetic_code=1,
        strand='both'):
    """
    ``tblastx`` similary search. Designed for using a translated
    nucleotide query on a translated nucleotide database.

    .. include:: blast_options_common.rst

    :param matrix: scoring matrix name. Available values are: PAM-30,
        PAM-70, BLOSUM-80, and BLOSUM-62.
    :param seg: filter query sequence with SEG as an integer (0 to
        disable, 1 to enable, or alternatively a tuple with the three
        parameters *window*, *locut*, and *hicut*).
    :param window_size: multiple hits window size (use 0 to specify
        1-hit algorithm).
    :param soft_masking: apply filtering locations as soft masks.
    :param lcase_masking: use lower case filtering in query and subject
        sequences.
    :param query_genetic_code: genetic code to translate query. Allowed
        values are: 1-6, 9-16, 21-25.
    :param db_genetic_code: genetic code to translate subject sequences.
        Allowed values are: 1-6, 9-16, 21-25.
    :param strand: query strand(s) to search against database/subject.
        Choice of both, minus, or plus.

    :return: A :class:`BlastOutput` instance.
    """

    # process common options
    cmd, stdin = _get_common_blast_options('tblastx', query, db, subject,
                        query_loc, subject_loc, evalue, False, num_threads)
    _get_common_protein_options(cmd, 'tblastx', word_size, matrix, None,
        None, None, seg, window_size, None)

    # filter/mask flags
    if soft_masking: cmd.extend(['-soft_masking', 'true'])
    else: cmd.extend(['-soft_masking', 'false'])
    if lcase_masking: cmd.append('-lcase_masking')

    # query_genetic_code
    if query_genetic_code not in [1, 2, 3, 4, 5, 6, 9, 10, 11, 12, 13, 14, 15, 16, 21, 22, 23, 24, 25]:
        raise ValueError('invalid genetic code: {0}'.format(query_genetic_code))
    cmd.extend(['-query_gencode', str(query_genetic_code)])

    # db_genetic_code
    if db_genetic_code not in [1, 2, 3, 4, 5, 6, 9, 10, 11, 12, 13, 14, 15, 16, 21, 22, 23, 24, 25]:
        raise ValueError('invalid genetic code: {0}'.format(db_genetic_code))
    cmd.extend(['-db_gencode', str(db_genetic_code)])

    # strand
    if not isinstance(strand, str): raise TypeError('`strand` must be a string')
    if strand not in ['both', 'minus', 'plus']: raise ValueError('`strand` must be either "both", "minus", or "plus"')
    cmd.extend(['-strand', strand])

    # run the program
    p = subprocess.Popen(cmd, stdin=None if stdin is None else subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    stdout, stderr = p.communicate(stdin)
    if p.returncode != 0:
        raise RuntimeError('error while running tblastx (message below)\n' + stderr)
    return _parse_blast_output(stdout)

def _get_common_blast_options(command, query, db, subject, query_loc,
                              subject_loc, evalue, parse_deflines,
                              num_threads):
    # get command name
    path = _blast_apps[command].get_path()
    if path is None:
        raise RuntimeError('{0} program not available -- please configure path'.format(command))
    cmd = [path, '-query', '-', '-outfmt', '5']

    # process query
    if isinstance(query, str):
        stdin = '>lcl|Query\n' + query + '\n'
        queryL = len(query)
    elif isinstance(query, _interface.SequenceView):
        if command in ['blastn', 'blastx', 'tblastx'] and query._parent._alphabet != alphabets.DNA: raise ValueError('query alphabet must be DNA')
        if command in ['blastp', 'tblastn'] and query._parent._alphabet != alphabets.protein: raise ValueError('query alphabet must be protein')
        stdin = '>lcl|Query\n' + query.string() + '\n'
        queryL = len(query)
    elif isinstance(query, _interface.SampleView):
        if command in ['blastn', 'blastx', 'tblastx'] and query._parent._alphabet != alphabets.DNA: raise ValueError('query alphabet must be DNA')
        if command in ['blastp', 'tblastn'] and query._parent._alphabet != alphabets.protein: raise ValueError('query alphabet must be protein')
        stdin = '>lcl|{0}\n{1}\n'.format(query.name, query.sequence.string())
        queryL = len(query.sequence)
    elif isinstance(query, (_interface.Align, _interface.Container)):
        if command in ['blastn', 'blastx', 'tblastx'] and query._alphabet != alphabets.DNA: raise ValueError('query alphabet must be DNA')
        if command in ['blastp', 'tblastn'] and query._alphabet != alphabets.protein: raise ValueError('query alphabet must be protein')
        stdin = query.fasta()
    else: raise TypeError('`query` must be an Align, a Container, a SampleView, a SequenceView, or a string')

    # process db/subject
    if db is not None and subject is not None:
        raise ValueError('only one of `db` and `subject` must be specified')
    if db is None and subject is None:
        raise ValueError('either `db` or `subject` must be specified')
    elif db is not None:
        if not isinstance(db, str): raise ValueError('`db` must be a string')
        if not os.path.isabs(db):
            db = os.path.normpath(os.path.join(_utils._protect_path_mapping[os.getcwd()], db))
        cmd.extend(['-db', db])
    else:
        f = open('in.fas', 'w')
        if isinstance(subject, str):
            f.write('>lcl|Subject\n{0}\n'.format(subject))
            subjectL = len(subject)
        elif isinstance(subject, _interface.SequenceView):
            if command in ['blastn', 'tblastn', 'tblastx'] and subject._parent._alphabet != alphabets.DNA: raise ValueError('subject alphabet must be DNA')
            if command in ['blastp', 'blastx'] and subject._parent._alphabet != alphabets.protein: raise ValueError('subject alphabet must be protein')
            f.write('>lcl|Subject\n{0}\n'.format(subject.string()))
            subjectL = len(subject)
        elif isinstance(subject, _interface.SampleView):
            if command in ['blastn', 'tblastn', 'tblastx'] and subject._parent._alphabet != alphabets.DNA: raise ValueError('subject alphabet must be DNA')
            if command in ['blastp', 'blastx'] and subject._parent._alphabet != alphabets.protein: raise ValueError('subject alphabet must be protein')
            f.write('>lcl|{0}\n{1}\n'.format(subject.name, subject.sequence.string()))
            subjectL = len(subject.sequence)
        else: raise TypeError('`subject` must be a SampleView, a SequenceView, or a string')
        cmd.extend(['-subject', 'in.fas'])

    # process query_loc
    if query_loc is not None:
        if isinstance(query, _interface.Container): raise ValueError('`query_loc` is not supported if `query` is a Container')
        if not isinstance(query_loc, (list, tuple)): raise TypeError('`query_loc` must be a list or a tuple')
        if len(query_loc) != 2: raise ValueError('`query_loc` must have two items')
        if not isinstance(query_loc[0], int) or not isinstance(query_loc[1], int): raise TypeError('`query_loc` must contain two integers')
        if query_loc[0] >= queryL or query_loc[0] < 0 or query_loc[1] < query_loc[0] + 1: raise ValueError('invalid values for `query_loc`')
        cmd.extend(['-query_loc', '{0}-{1}'.format(query_loc[0]+1, query_loc[1])])

    # process subject_loc
    if subject_loc is not None:
        if subject is None: raise ValueError('`subject_loc` cannot be specified if `subject` is not specified')
        if not isinstance(subject_loc, (list, tuple)): raise TypeError('`subject_loc` must be a list or a tuple')
        if len(subject_loc) != 2: raise ValueError('`subject_loc` must have two items')
        if not isinstance(subject_loc[0], int) or not isinstance(subject_loc[1], int): raise TypeError('`subject_loc` must contain two integers')
        if subject_loc[0] >= subjectL or subject_loc[0] < 0 or subject_loc[1] < subject_loc[0] + 1: raise ValueError('invalid values for `subject_loc`')
        cmd.extend(['-subject_loc', '{0}-{1}'.format(subject_loc[0]+1, subject_loc[1])])

    # process evalue
    if evalue is not None:
        if not isinstance(evalue, (int, float)): raise TypeError('`evalue` must be a real value')
        if evalue <= 0: raise ValueError('`evalue` must be > 0')
        cmd.extend(['-evalue', str(evalue)])

    # parse_deflines
    if parse_deflines: cmd.append('-parse_deflines')

    # num_threads
    if not isinstance(num_threads, int): raise TypeError('`num_threads` must be an integer')
    if num_threads < 1: raise ValueError('`num_threads` must be >= 1')
    if num_threads > 1 and subject is not None: raise ValueError('`num_threads cannot be > 1 if `subject` is used')
    cmd.extend(['-num_threads', str(num_threads)])

    return cmd, stdin

def _get_common_protein_options(cmd, task, word_size, matrix, threshold,
        gapopen, gapextend, seg, window_size, comp_based_stats):

    # word size
    if word_size is not None:
        if not isinstance(word_size, int): raise TypeError('`word_size` must be an integer')
        if word_size < 2 or word_size > 7: raise ValueError('invalid value for `word_size`: must be in range [2,7]')
        cmd.extend(['-word_size', str(word_size)])

    # matrix
    if matrix is not None:
        if matrix not in ['PAM30', 'PAM70', 'BLOSUM80', 'BLOSUM62']:
            raise ValueError('invalid value for `matrix`: {0}'.format(matrix))
        cmd.extend(['-matrix', matrix])

    # gap costs
    if task not in ['blastp-fast', 'tblastx']:
        if gapopen is None:
            if gapextend is not None: raise ValueError('if `gapopen is ``None``, `gapextend` must be also')
            gapopen, gapextend = _blastp_costs_defaults[matrix]
        else:
            if gapextend is None: raise ValueError('if `gapextend is ``None``, `gapopen` must be also')
            if not isinstance(gapopen, int): raise TypeError('`gapopen` must be an integer')
            if not isinstance(gapextend, int): raise TypeError('`gapextend` must be an integer')
            if (gapopen, gapextend) not in _blastp_costs[matrix]: raise ValueError('invalid gapopen/gapextend combination for {0}'.format(matrix))
        cmd.extend(['-gapopen', str(gapopen), '-gapextend', str(gapextend)])

    # threshold
    if threshold is not None:
        if not isinstance(threshold, int): raise TypeError('`threshold` must be an integer')
        if threshold <= 0: raise ValueError('invalid value for `treshold`: must be > 0')
        cmd.extend(['-threshold', str(threshold)])

    # seg
    if seg is not None:
        if seg == 0: seg = 'no'
        elif seg == 1: seg = 'yes'
        elif not isinstance(seg, (tuple, list)):
            raise ValueError('invalid value for `seg`: must be 0, 1, or a tuple/list of 3 items')
        else:
            try:
                window, locut, hicut = seg
                window = int(window)
            except ValueError:
                raise ValueError('invalid value for `seg`')
            if window < 1: raise ValueError('`window` (seg parameter) must be > 0')
            locut = float(locut)
            if locut <= 0: raise ValueError('`locut` (seg parameter) must be > 0')
            hicut = float(hicut)
            if hicut <= locut: raise ValueError('`hicut` (seg parameter) must be > `locut`')
            seg = '{0} {1} {2}'.format(window, locut, hicut)
        cmd.extend(['-seg', seg])

    # window_size
    if window_size is not None:
        if not isinstance(window_size, int): raise TypeError('`window_size` must be an integer')
        if window_size < 0: raise ValueError('invalid value for `window_size`: {0}'.format(window_size))
        cmd.extend(['-window_size', str(window_size)])

    # comp_based_stats
    if comp_based_stats is not None:
        if comp_based_stats not in [0, 1, 2, 3]:
            raise ValueError('invalid value for `comp_based_stats`')
        cmd.extend(['-comp_based_stats', str(comp_based_stats)])

def _xml_find(node, tag):
    res = node.find(tag)
    if res is None: raise ValueError('invalid format of BLAST XML output: cannot find `{0}` under `{1}`'.format(tag, node.tag))
    return res

def _xml_find_type(node, tag, type_):
    value = _xml_find(node, tag).text
    try: return type_(value)
    except ValueError:
        raise ValueError('invalid format of BLAST XML output: invalid value for `{0}`: {1}'.format(tag, value))

def _parse_blast_output(string):
    root = xml.etree.ElementTree.fromstring(string)
    if root.tag != 'BlastOutput': raise ValueError('invalid format of BLAST XML output: no `BlastOutput` at root')

    out = BlastOutput()
    out._program = _xml_find(root, 'BlastOutput_program').text
    out._version = _xml_find(root, 'BlastOutput_version').text
    out._reference = _xml_find(root, 'BlastOutput_reference').text
    out._db = _xml_find(root, 'BlastOutput_db').text
    out._query_ID = _xml_find(root, 'BlastOutput_query-ID').text
    out._query_def = _xml_find(root, 'BlastOutput_query-def').text
    out._query_len = _xml_find_type(root, 'BlastOutput_query-len',  int)

    node = _xml_find(root, 'BlastOutput_param')
    nodes = node.findall('Parameters')
    if len(nodes) != 1: raise ValueError('invalid format of BLAST XML output: expect one set of BLAST parameters')
    node = nodes[0]

    out._params = {}
    for child in node:
        if child.tag[:11] != 'Parameters_':
            raise ValueError('invalid parameter in BLAST output: {0}'.format(child.tag))
        tag = child.tag[11:]
        if tag in out._params:
            raise ValueError('parameter repeated in BLAST output: {0}'.format(child.tag))
        try: out._params[tag] = int(child.text)
        except ValueError:
            try: out._params[tag] = float(child.text)
            except ValueError:
                out._params[tag] = child.text

    iterations = _xml_find(root, 'BlastOutput_iterations')
    for query_node in iterations:
        if query_node.tag != 'Iteration': raise ValueError('invalid format of BLAST XML output: unexpected node under `BlastOutput_iterations`: {0}'.format(query_node.tag))
        query_hits = BlastQueryHits()
        i = _xml_find_type(query_node, 'Iteration_iter-num', int)
        if i != len(out._query_hits) + 1: raise ValueError('invalid format of BLAST XML output: unexpected value for `Iteration_iter-num`: {0}'.format(i))
        query_hits._num = i - 1
        query_hits._query_ID = _xml_find(query_node, 'Iteration_query-ID').text
        query_hits._query_def = _xml_find(query_node, 'Iteration_query-def').text
        query_hits._query_len = _xml_find_type(query_node, 'Iteration_query-len', int)
        node = _xml_find(query_node, 'Iteration_stat')
        nodes = node.findall('Statistics')
        if len(nodes) != 1: raise ValueError('invalid format of BLAST XML output: expect one set of iteration statistics')
        node = nodes[0]
        query_hits._db_num = _xml_find_type(node, 'Statistics_db-num', int)
        query_hits._db_len = _xml_find_type(node, 'Statistics_db-len', int)
        query_hits._hsp_len = _xml_find_type(node, 'Statistics_hsp-len', int)
        query_hits._eff_space = _xml_find_type(node, 'Statistics_eff-space', int)
        query_hits._kappa = _xml_find_type(node, 'Statistics_kappa', float)
        query_hits._lambda = _xml_find_type(node, 'Statistics_lambda', float)
        query_hits._entropy = _xml_find_type(node, 'Statistics_entropy', float)
        out._query_hits.append(query_hits)

        hits = _xml_find(query_node, 'Iteration_hits')
        for hit_node in hits:
            if hit_node.tag != 'Hit': raise ValueError('invalid format of BLAST XML output: unexpected node under `Iteration_hits`: {0}'.format(hit_node.tag))
            hit = BlastHit()
            i = _xml_find_type(hit_node, 'Hit_num', int)
            if i != len(query_hits._hits) + 1: raise ValueError('invalid format of BLAST XML output: unexpected value for `Hit_num`: {0}'.format(i))
            hit._num = i - 1
            hit._id = _xml_find(hit_node, 'Hit_id').text
            hit._def = _xml_find(hit_node, 'Hit_def').text
            hit._accession = _xml_find(hit_node, 'Hit_accession').text
            hit._len = _xml_find_type(hit_node, 'Hit_len', int)
            query_hits._hits.append(hit)

            Hsps = _xml_find(hit_node, 'Hit_hsps')
            for Hsp_node in Hsps:
                if Hsp_node.tag != 'Hsp': raise ValueError('invalid format of BLAST XML output: unexpected node under `Hit_hsps`: {0}'.format(Hsp_node.tag))
                Hsp = BlastHsp()
                i = _xml_find_type(Hsp_node, 'Hsp_num', int)
                if i != len(hit._hsp) + 1: raise ValueError('invalid format of BLAST XML output: unexpected value for `Hsp_num`: {0}'.format(i))
                Hsp._num = i - 1
                Hsp._bit_score = _xml_find_type(Hsp_node, 'Hsp_bit-score', float)
                Hsp._evalue = _xml_find_type(Hsp_node, 'Hsp_evalue', float)
                Hsp._identity = _xml_find_type(Hsp_node, 'Hsp_identity', int)
                Hsp._gaps = _xml_find_type(Hsp_node, 'Hsp_gaps', int)
                Hsp._positive = _xml_find_type(Hsp_node, 'Hsp_positive', int)
                Hsp._align_len = _xml_find_type(Hsp_node, 'Hsp_align-len', int)
                Hsp._qseq = _xml_find(Hsp_node, 'Hsp_qseq').text
                Hsp._midline = _xml_find(Hsp_node, 'Hsp_midline').text
                Hsp._hseq = _xml_find(Hsp_node, 'Hsp_hseq').text
                if len(Hsp._qseq) != Hsp._align_len: raise ValueError('invalid format of BLAST XML output: incompatibility between `qseq` and `align-len`')
                if len(Hsp._hseq) != Hsp._align_len: raise ValueError('invalid format of BLAST XML output: incompatibility between `hseq` and `align-len`')
                if len(Hsp._midline) != Hsp._align_len: raise ValueError('invalid format of BLAST XML output: incompatibility between `midline` and `align-len`')

                qfrom = _xml_find_type(Hsp_node, 'Hsp_query-from', int)
                qto = _xml_find_type(Hsp_node, 'Hsp_query-to', int)
                Hsp._query_frame = _xml_find_type(Hsp_node, 'Hsp_query-frame', int)
                if out._program in ['blastp', 'tblastn']:
                    if Hsp._query_frame != 0:  raise ValueError('invalid value for `Hsp_query-frame`: {0}'.format(Hsp._query_frame))
                    if qto <= qfrom: raise ValueError('invalid values for `Hsp_query-from/to`: {0} and {1} with positive frame'.format(qfrom, qto))
                    Hsp._query_start = qfrom - 1
                    Hsp._query_stop = qto
                else:
                    if Hsp._query_frame not in [-3, -2, -1, 1, 2, 3]: raise ValueError('invalid value for `Hsp_query-frame`: {0}'.format(Hsp._query_frame))
                    qfrom, qto = sorted([qfrom, qto])
                    Hsp._query_start = qfrom - 1
                    Hsp._query_stop = qto

                hfrom = _xml_find_type(Hsp_node, 'Hsp_hit-from', int)
                hto = _xml_find_type(Hsp_node, 'Hsp_hit-to', int)
                Hsp._hit_frame = _xml_find_type(Hsp_node, 'Hsp_hit-frame', int)
                if out._program in ['blastp', 'blastx']:
                    if Hsp._hit_frame != 0:  raise ValueError('invalid value for `Hsp_hit-frame`: {0}'.format(Hsp._hit_frame))
                    if hto <= hfrom: raise ValueError('invalid values for `Hsp_hit-from/to`: {0} and {1} with positive frame'.format(hfrom, hto))
                    Hsp._hit_start = hfrom - 1
                    Hsp._hit_stop = hto
                else:
                    if Hsp._hit_frame not in [-3, -2, -1, 1, 2, 3]: raise ValueError('invalid value for `Hsp_hit-frame`: {0}'.format(Hsp._hit_frame))
                    hfrom, hto = sorted([hfrom, hto])
                    Hsp._hit_start = hfrom - 1
                    Hsp._hit_stop = hto

                hit._hsp.append(Hsp)
    return out

class BlastOutput(object):
    """
    Full results of a BLAST run.
    """
    def __init__(self):
        self._program = None
        self._version = None
        self._reference = None
        self._db = None
        self._query_ID = None
        self._query_def = None
        self._query_len = None
        self._params = None
        self._query_hits = []

    @property
    def program(self):
        """ Name of the program used."""
        return self._program

    @property
    def version(self):
        """ Version of the program. """
        return self._version

    @property
    def reference(self):
        """ Bibliographic reference. """
        return self._reference

    @property
    def db(self):
        """ Name of the database used. """
        return self._db

    @property
    def query_ID(self):
        """ Identifier of the query. """
        return self._query_ID

    @property
    def query_def(self):
        """ Description of the query. """
        return self._query_def

    @property
    def query_len(self):
        """ Length of the query. """
        return self._query_len

    @property
    def params(self):
        """
        Search parameters.
        
        ``"expect"``: E-value, ``"reward"``:
        nucleotide match reward, ``"penalty"``: nucleotide
        mismatch reward, ``"gapopen"``: cost for opening a gap,
        ``"gapextend"``: cost for extending a gap.
        ``"filter"``: filter string.
        """
        return self._params

    @property
    def num_queries(self):
        """
        Number of queries used in the BLAST search. ``blast_output.num_queries``
        is also available as ``len(blast_output)``.
        """
        return len(self._query_hits)

    def __len__(self):
        return len(self._query_hits)

    def iter_queries(self):
        """
        Iterator over queries. Allows to iterate over
        :class:`.BlastQueryHits` instances for all queries.
        ``for query_hit in blast_output.iter_queries()`` is
        also available as ``for query_hit in blast_output``.
        """
        return self.__iter__()

    def __iter__(self):
        for i in self._query_hits: yield i

    def get_query(self, i):
        """
        Hits for a given query. An instance of :class:`.BlastQueryHits`
        is returned.
        ``blast_output.get_query(i)`` is also available as ``blast_output[i]``.
        """
        return self._query_hits[i]

    def __getitem__(self, i):
        return self._query_hits[i]

    @property
    def num_hits(self):
        """
        Total number of hits for all queries.
        """
        return sum(map(len, self._query_hits))

    def iter_hits(self):
        """
        Iterator over all hits of all queries. Iterates over
        :class:`.BlastHit` instances
        """
        for query in self._query_hits:
            for hit in query:
                yield hit

    @property
    def num_hsp(self):
        """
        Total number of Hsp's for all hits of all entries.
        """
        return sum([hit.num_hsp() for hit in self._query_hits])

    def iter_hsp(self):
        """
        Iterator over all Hsp's of all hits of all queries.
        """
        for query in self._query_hits:
            for hit in query:
                for Hsp in hit:
                    yield Hsp

class BlastQueryHits(object):
    """
    Results for a given query of a BLAST run.
    """
    def __init__(self):
        self._num = None
        self._query_ID = None
        self._query_def = None
        self._query_len = None
        self._db_num = None
        self._db_len = None
        self._hsp_len = None
        self._eff_space = None
        self._kappa = None
        self._lambda = None
        self._entropy = None
        self._hits = []

    @property
    def num(self):
        """ index of the query in the BLAST run. """
        return self._num

    @property
    def query_ID(self):
        """Identifier of the query."""
        return self._query_ID

    @property
    def query_len(self):
        """Length of the query."""
        return self._query_len

    @property
    def query_def(self):
        """Description of the query."""
        return self._query_def

    @property
    def db_num(self):
        """Number of sequence in the database."""
        return self._db_num

    @property
    def db_len(self):
        """Number of letters in the database."""
        return self._db_len

    @property
    def hsp_len(self):
        """Length adjustment."""
        return self._hsp_len

    @property
    def eff_space(self):
        """Effective space of the search."""
        return self._eff_space

    @property
    def K(self):
        """Karlin-Altschul kappa parameter."""
        return self._kappa

    @property
    def L(self):
        """Karlin-Altschul lambda parameter."""
        return self._lambda

    @property
    def H(self):
        """Karlin-Altschul entropy parameter."""
        return self._entropy

    @property
    def num_hits(self):
        """
        Number of hits for this query. ``query_hits.num_hits()`` is also
        available as ``len(query_hits)``.
        """
        return len(self._hits)

    def __len__(self):
        return len(self._hits)

    def iter_hits(self):
        """
        Iterator to the :class:`.BlastHit` instances of all
        hits. ``for hit in query_hits.iter_hits()`` is also available as
        ``for hit in query_hits``.
        """
        for i in self._hits: yield i

    def __iter__(self):
        for i in self._hits: yield i

    def get_hit(self, i):
        """
        Get a given hit, as a :class:`.BlastHit` instance. ``query_hits.get_hit(i)``
        is also available as ``query_hits[i]``.
        """
        return self._hits[i];

    def __getitem__(self, i):
        return self._hits[i]

    @property
    def num_hsp(self):
        """
        Total number of Hsp's for all hits.
        """
        return sum(map(len, self._hits))

    def iter_hsp(self):
        """
        Iterator over all Hsp's of all hits, as :class:`.BlastHsp`
        instances
        """
        for hit in self._hits:
            for hsp in hit:
                yield hsp

class BlastHit(object):
    """
    Results for a given hit of a BLAST run.
    """
    def __init__(self):
        self._num = None
        self._id = None
        self._def = None
        self._accession = None
        self._len = None
        self._hsp = []

    @property
    def num(self):
        """Index of the hit for the corresponding query."""
        return self._num

    @property
    def id(self):
        """Identifier of the subject."""
        return self._id

    @property
    def descr(self):
        """Description of the subject."""
        return self._def

    @property
    def accession(self):
        """Identifier of the subject."""
        return self._accession

    @property
    def len(self):
        """Length of subject."""
        return self._len

    @property
    def num_hsp(self):
        """
        Number of Hsp's in this hit. ``hit.num_Hsp()  is also available as
        ``len(hit)``.
        """
        return len(self._hsp)

    def __len__(self):
        return len(self._hsp)

    def iter_hsp(self):
        """
        Iterator to the :class:`.BlastHsp` instances for all
        Hsp's. ``for hsp in hit.iter_hsp()`` is also available as
        ``for hsp in hit``.
        """
        for i in self._hsp: yield i

    def __iter__(self):
        for i in self._hsp: yield i

    def get_hsp(self, i):
        """
        Get a given Hsp, as a :class:`.BlastHsp` instance. ``hit.get_hsp(i)``
        is also available as ``hit[i]``.
        """
        return self._hsp[i];

    def __getitem__(self, i):
        return self._hsp[i]

class BlastHsp(object):
    """
    Description of an Hsp of a BLAST run.

    Start and stop positions are always interpreted as range parameters (use
    `frame` to determine if the complement should be used)::

        >>> hit_sequence = seq[query_start:query_to]
    """
    def __init__(self):
        self._num = None
        self._bit_score = None
        self._evalue = None
        self._query_start = None
        self._query_stop = None
        self._query_frame = None
        self._hit_start = None
        self._hit_stop = None
        self._hit_frame = None
        self._identity = None
        self._positive = None
        self._gaps = None
        self._align_len = None
        self._qseq = None
        self._hseq = None
        self._midline = None

    @property
    def num(self):
        """Index of the Hsp in the corresponding hit."""
        return self._num

    @property
    def bit_score(self):
        """Bit score of the Hsp."""
        return self._bit_score

    @property
    def evalue(self):
        """Expectation value of the Hsp."""
        return self._evalue

    @property
    def query_start(self):
        """Start position on the query."""
        return self._query_start

    @property
    def query_stop(self):
        """Stop position on the query."""
        return self._query_stop

    @property
    def hit_start(self):
        """Start position on the subject."""
        return self._hit_start

    @property
    def hit_stop(self):
        """Stop position on the subject."""
        return self._hit_stop

    @property
    def query_frame(self):
        """Frame of the query."""
        return self._query_frame

    @property
    def hit_frame(self):
        """Frame of the hit."""
        return self._hit_frame

    @property
    def identity(self):
        """Number of identical positions."""
        return self._identity

    @property
    def positive(self):
        """Number of positions with positive score."""
        return self._positive

    @property
    def gaps(self):
        """Number of gap positions."""
        return self._gaps

    @property
    def align_len(self):
        """Length of the alignment."""
        return self._align_len

    @property
    def qseq(self):
        """Aligned query sequence."""
        return self._qseq

    @property
    def hseq(self):
        """Aligned subject sequence."""
        return self._hseq

    @property
    def midline(self):
        """Alignment midline."""
        return self._midline



"""
    Copyright 2016-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

from . import eggwrapper as _eggwrapper
from . import _interface

def site_from_align(align, index):
    """
    Create a site from a position of an alignment.
    Import allelic and genotypic data from a position of the provided
    :class:`.Align` instance.

    :param align: an :class:`!Align` instance.
    :param index: the index of a valid (not out of range) position of
        the alignment.
    :return: A new :class:`!Site` instance.
    """
    obj = Site()
    obj.from_align(align, index)
    return obj

def site_from_list(array, alphabet):
    """
    Create a site based on a list of allelic values.

    :param array: a :class:`list` of allelic values.
    :param alphabet: an alphabet (input allelic values must match this
        alphabet).
    """
    obj = Site()
    obj.from_list(array, alphabet)
    return obj

def site_from_vcf(vcf, start=0, stop=None):
    """
    Extract a site from a VCF parser.
    Import allelic and genotypic data from a VCF parser. The VCF parser
    must have processed a variant and the variant is required to have
    genotypic data available as the GT format field. An exception is
    raised otherwise.

    :param vcf: an :class:`.io.VcfParser` instance containing data. There
        must at least one sample and the GT format field must be
        available. It is not required to extract variant data manually.

    :param start: index of the first sample to process. Index is
        required to be within available bounds (it must be at least 0
        and smaller than the number of samples in the VCF data). Note
        that in a VCF file a sample corresponds to a genotype.

    :param stop: sample index at which processing must be stopped (this
        sample is not processed). Index is required to be within
        available bounds (if must be at least equal to *start* and not
        larger than the number of samples in the VCF data). Note that in
        a VCF file, a sample corresponds to a genotype.
    """
    obj = Site()
    obj.from_vcf(vcf, start, stop)
    return obj

class Site(object):
    """
    Store alleles from a single site. Instances can be created from a
    position in an :class:`.Align` instance (using :func:`.site_from_align`),
    from the current position of a :class:`.VcfParser` (using :func:`.site_from_vcf`),
    or from a user-provided list of allelic values (using :func:`.site_from_list`).

    :param alphabet: an :class:`.Alphabet` instance (only useful if
        data are supposed to be loaded one by one :meth:`~.Site.append`
        or :meth:`~.Site.extend`).

    The following operations are available on ``site`` if it is a :class:`!Site`
    instance:

    +------------------------+---------------------------------+
    | Operation              | Result                          |
    +========================+=================================+
    | ``len(site)``          | number of samples               |
    +------------------------+---------------------------------+
    | ``for i in site: ...`` | iterate over alleles            |
    +------------------------+---------------------------------+
    | ``site[i]``            | access allele at given index    |
    +------------------------+---------------------------------+
    | ``site[i] = a``        | overwrite allele at given index |
    +------------------------+---------------------------------+
    | ``del site[i]``        | delete allele at given index    |
    +------------------------+---------------------------------+
    """

    @classmethod
    def _from_site_holder(cls, obj, alphabet):
        ret = object.__new__(cls)
        ret._obj = obj
        ret._alphabet = alphabet
        return ret

    def __init__(self, alphabet=None):
        self._obj = _eggwrapper.SiteHolder()
        self._alphabet = alphabet

    def reset(self):
        """Clear all data from the instance (including the alphabet)."""
        self._obj.reset()
        self._alphabet = None

    @property
    def position(self):
        """
        Position of the site. The position is set automatically if the
        instance is created or reset from an :class:`.Align` or a
        :class:`.VcfParser`. In all cases, the value can be modified.
        """
        return None if self._obj.get_position() == _eggwrapper.UNDEF else self._obj.get_position()

    @position.setter
    def position(self, value):
        self._obj.set_position(value)

    @property
    def ns(self):
        """Number of samples"""
        return self._obj.get_ns()

    def __len__(self):
        return self._obj.get_ns()

    @property
    def num_missing(self):
        """Number of missing data."""
        return self._obj.get_missing()

    @property
    def alphabet(self):
        """
        Alphabet attached to the instance. It is possible to set the
        alphabet, but not change it (the alphabet can be changed only
        immediately after creation if no alphabet has been
        specified, or after calling :meth:`~.Site.reset`.
        """
        return self._alphabet

    @alphabet.setter
    def alphabet(self, alph):
        if self._alphabet is not None: raise ValueError('cannot change alphabet of a Site')
        self._alphabet = alph

    def as_list(self):
        """
        Generate a list containing data from the instance.

        :return: A :class:`list` of allelic values.
        """
        return [self._alphabet.get_value(self._obj.get_sample(i)) for i in range(self._obj.get_ns())]

    def __getitem__(self, idx):
        if isinstance(idx, slice): raise ValueError('slices are not supported')
        if idx < 0: idx = self._obj.get_ns() + idx
        if idx < 0 or idx >= self._obj.get_ns(): raise IndexError('invalid sample index')
        return self._alphabet.get_value(self._obj.get_sample(idx))

    def __setitem__(self, idx, val):
        if isinstance(idx, slice): raise ValueError('slices are not supported')
        if idx < 0: idx = self._obj.get_ns() + idx
        if idx < 0 or idx >= self._obj.get_ns(): raise IndexError('invalid sample index')
        self._obj.set_sample(idx, self._alphabet.get_code(val))

    def __delitem__(self, idx):
        if isinstance(idx, slice): raise ValueError('slices are not supported')
        if idx < 0: idx = self._obj.get_ns() + idx
        if idx < 0 or idx >= self._obj.get_ns(): raise IndexError('invalid sample index')
        self._obj.del_sample(idx)

    def append(self, val):
        """ Add an allele at the end of the site. """
        self._obj.add(1)
        self._obj.set_sample(self._obj.get_ns()-1, self._alphabet.get_code(val))

    def extend(self, vals):
        """
        Add several alleles at the end of the site. The alleles
        must be provided as an iterable. """
        old_ns = self._obj.get_ns()
        self._obj.add(len(vals))
        for i, val in enumerate(vals):
            self._obj.set_sample(old_ns+i, self._alphabet.get_code(val))

    def from_align(self, align, index):
        """
        Import data from the provided :class:`.Align`. All data
        currently stored in the instance are discarded.

        :param align: an :class:`!Align` instance.
        :param index: index of the site to extract.
        """
        self._obj.reset()
        self._alphabet = align._alphabet
        if index >= align.ls: raise IndexError('invalid site index')
        self._obj.process_align(align._obj, index)
        self._position = index

    def from_list(self, array, alphabet):
        """
        Import alleles from the provided list.

        :param array: a :class:`list` of allelic values.
        :param alphabet: an alphabet (input allelic values must match this
            alphabet).
        """
        self._obj.reset()
        self._alphabet = alphabet
        if len(array) == 0: return
        self._obj.add(len(array))
        for i, v in enumerate(array):
            self._obj.set_sample(i, self._alphabet.get_code(v))
        self._position = None

    def from_vcf(self, vcf, start=0, stop=None):
        """
        Import data from the provided :class:`.io.VcfParser`.  
        The VCF parser
        must have processed a variant and the variant is required to have
        genotypic data available as the GT format field. An exception is
        raised otherwise.

        :param vcf: a :class:`!VcfParser` instance containing data. There
            must at least one sample and the GT format field must be
            available. It is not required to extract variant data manually.

        :param start: index of the first sample to process. Index is
            required to be within available bounds (it must be at least 0
            and smaller than the number of samples in the VCF data). Note
            that in a VCF file a sample corresponds to a genotype.

        :param stop: sample index at which processing must be stopped (this
            sample is not processed). Index is required to be within
            available bounds (if must be at least equal to *start* and not
            larger than the number of samples in the VCF data). Note that in
            a VCF file, a sample corresponds to a genotype.
        """
        self._obj.reset()
        vcf.get_genotypes(start, stop, self)



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import egglib, unittest, pathlib
path = pathlib.Path(__file__).parent / '..' / 'data'

class CodingDiversity_test(unittest.TestCase):
    def test_CodingDiversity_T(self):
        cdiv = egglib.stats.CodingDiversity()
        self.assertEqual(str(type(cdiv)), "<class 'egglib.stats._coding_diversity.CodingDiversity'>")

    def test_process_T(self):
        cdiv= egglib.stats.CodingDiversity()
        frame_aln = egglib.tools.ReadingFrame([[0, 698, 3], [1264, 1318, None], [1403, 1513, None], [3431, 3663, None], [4267, 4474, None], [5733, 5792, None], [5882, 6039, None]])
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        aln.to_codons(frame=frame_aln)
        v_a=cdiv.num_codons_tot
        cdiv.process(aln, struct=egglib.struct_from_labels(aln), max_missing=0)
        v_b=cdiv.num_codons_tot
        self.assertTrue(v_b>v_a)
        self.assertEqual(str(type(cdiv)), "<class 'egglib.stats._coding_diversity.CodingDiversity'>")

    def test_process_E(self):
        cdiv= egglib.stats.CodingDiversity()
        frame_aln = egglib.tools.ReadingFrame([[0, 698, 3], [1264, 1318, None], [1403, 1513, None], [3431, 3663, None], [4267, 4474, None], [5733, 5792, None], [5882, 6039, None]])
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        aln.to_codons(frame=frame_aln)
        with self.assertRaises(ValueError):
            cdiv.process(aln, struct=egglib.struct_from_labels(aln), max_missing=0, code=1000)
        cdiv.process(aln, max_missing=0, code=1)

    def test_num_codons_tot_T(self):
        cdiv= egglib.stats.CodingDiversity()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        frame = egglib.tools.ReadingFrame([[0, 6039]])
        aln.to_codons(frame)
        cdiv.process(aln, struct=egglib.struct_from_labels(aln), max_missing=0)
        nct=cdiv.num_codons_tot
        self.assertEqual(nct, 2013)

    def test_num_codons_eff_T(self):
        cdiv= egglib.stats.CodingDiversity()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        frame = egglib.tools.ReadingFrame([[0, 6039]])
        aln.to_codons(frame)
        cdiv.process(aln, struct=egglib.struct_from_labels(aln), max_missing=0)
        nce=cdiv.num_codons_eff
        self.assertEqual(nce, 472)

    def test_num_codons_stop_T(self):
        cdiv= egglib.stats.CodingDiversity()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        frame = egglib.tools.ReadingFrame([[0, 698, 3], [1264, 1318, None], [1403, 1513, None], [3431, 3663, None], [4267, 4474, None], [5733, 5792, None], [5882, 6039, None]])
        aln.to_codons(frame=frame)
        cdiv.process(aln, struct=egglib.struct_from_labels(aln), max_missing=0)
        ncs=cdiv.num_codons_stop
        self.assertEqual(ncs, 14)

    def test_num_sites_NS_T(self):
        cdiv= egglib.stats.CodingDiversity()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        frame = egglib.tools.ReadingFrame([[0, 6039]])
        aln.to_codons(frame)
        cdiv.process(aln, struct=egglib.struct_from_labels(aln), max_missing=0)
        nsN=cdiv.num_sites_NS
        self.assertEqual(round(nsN,8),1133.14327485)

    def test_num_sites_S_T(self):
        cdiv= egglib.stats.CodingDiversity()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        frame = egglib.tools.ReadingFrame([[0, 6039]])
        aln.to_codons(frame)
        cdiv.process(aln, struct=egglib.struct_from_labels(aln), max_missing=0)
        nsS=cdiv.num_sites_S
        self.assertEqual(round(nsS,9),282.856725146)

    def test_num_pol_single_T(self):
        cdiv= egglib.stats.CodingDiversity()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        frame = egglib.tools.ReadingFrame([[0, 6039]])
        aln.to_codons(frame)
        cdiv.process(aln, struct=egglib.struct_from_labels(aln), max_missing=0)
        nps=cdiv.num_pol_single
        self.assertEqual(nps, 37)

    def test_num_pol_multi_T(self):
        cdiv= egglib.stats.CodingDiversity()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        frame = egglib.tools.ReadingFrame([[0, 698, 3], [1264, 1318, None], [1403, 1513, None], [3431, 3663, None], [4267, 4474, None], [5733, 5792, None], [5882, 6039, None]])
        aln.to_codons(frame=frame)
        cdiv.process(aln, struct=egglib.struct_from_labels(aln), max_missing=999)
        npm=cdiv.num_multiple_hits
        self.assertEqual(npm, 2)

    def test_num_pol_NS_T(self):
        cdiv= egglib.stats.CodingDiversity()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        frame = egglib.tools.ReadingFrame([[0, 698, 3], [1264, 1318, None], [1403, 1513, None], [3431, 3663, None], [4267, 4474, None], [5733, 5792, None], [5882, 6039, None]])
        aln.to_codons(frame=frame)
        cdiv.process(aln, struct=egglib.struct_from_labels(aln), max_missing=999)
        npNS=cdiv.num_pol_NS
        self.assertEqual(npNS,5)

    def test_num_pol_S_T(self):
        cdiv= egglib.stats.CodingDiversity()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        frame = egglib.tools.ReadingFrame([[0, 698, 3], [1264, 1318, None], [1403, 1513, None], [3431, 3663, None], [4267, 4474, None], [5733, 5792, None], [5882, 6039, None]])
        aln.to_codons(frame=frame)
        cdiv.process(aln, struct=egglib.struct_from_labels(aln), max_missing=999)
        npS=cdiv.num_pol_S
        self.assertEqual(npS, 16)

    def test_iter_S_T(self):
        cdiv= egglib.stats.CodingDiversity()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        frame = egglib.tools.ReadingFrame([[0, 698, 3], [1264, 1318, None], [1403, 1513, None], [3431, 3663, None], [4267, 4474, None], [5733, 5792, None], [5882, 6039, None]])
        aln.to_codons(frame=frame)
        cdiv.process(aln, struct=egglib.struct_from_labels(aln), max_missing=999)
        self.assertIsInstance(cdiv.sites_S, list)
        self.assertEqual(len(cdiv.sites_S), 16)

    def test_iter_NS_T(self):
        cdiv= egglib.stats.CodingDiversity()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        frame = egglib.tools.ReadingFrame([[0, 698, 3], [1264, 1318, None], [1403, 1513, None], [3431, 3663, None], [4267, 4474, None], [5733, 5792, None], [5882, 6039, None]])
        aln.to_codons(frame=frame)
        cdiv.process(aln, struct=egglib.struct_from_labels(aln), max_missing=999)
        self.assertIsInstance(cdiv.sites_NS, list)
        self.assertEqual(len(cdiv.sites_NS), 5)

    def test_codon_diversity(self):
        aln = egglib.Align.create([
        # SYN              +                    +
        # NON-SYN                 +      +       
            ('', ['ATG', 'ACG', 'TAC', 'GAT', 'CCC', 'TAA']),
            ('', ['ATG', 'ACA', 'TAC', 'GTT', 'CCC', 'TAA']),
            ('', ['ATG', 'ACA', 'TCC', 'GTT', 'CCC', 'TAA']),
            ('', ['AT?', 'ACA', 'TCC', 'GAT', 'CCA', 'TAA'])], egglib.alphabets.codons)

        cd = egglib.stats.CodingDiversity(aln)
        cs = egglib.stats.ComputeStats()
        cs.add_stats('Ki', 'D', 'S')
        statsS = cs.process_sites(cd.sites_S)
        statsNS = cs.process_sites(cd.sites_NS)
        self.assertEqual(statsS['S'], 2)
        self.assertEqual(statsNS['S'], 2)
        self.assertEqual(statsS['Ki'], 3)
        self.assertEqual(statsNS['Ki'], 4)

    def test_raise_stop(self):
        aln = egglib.Align.create([
        # SYN              +                    +
        # NON-SYN                 +      +       
        # STOP                                         +
            ('', ['ATG', 'ACG', 'TAC', 'GAT', 'CCC', 'TCA']),
            ('', ['ATG', 'ACA', 'TAC', 'GTT', 'CCC', 'TCA']),
            ('', ['ATG', 'ACA', 'TCC', 'GTT', 'CCC', 'TCA']),
            ('', ['AT?', 'ACA', 'TCC', 'GAT', 'CCA', 'TAA'])], egglib.alphabets.codons)

        with self.assertRaises(ValueError):
            cd = egglib.stats.CodingDiversity(aln, raise_stop=True, skipstop=True)

        with self.assertRaises(ValueError):
            cd = egglib.stats.CodingDiversity(aln, raise_stop=True, skipstop=False)

        cd = egglib.stats.CodingDiversity(aln, skipstop=True)
        self.assertListEqual(cd.positions_S, [1, 4])
        self.assertListEqual(cd.positions_NS, [2, 3])

        cd = egglib.stats.CodingDiversity(aln, skipstop=False)
        self.assertListEqual(cd.positions_S, [1, 4])
        self.assertListEqual(cd.positions_NS, [2, 3, 5])



"""
    Copyright 2016-2021 Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

from .. import eggwrapper as _eggwrapper
from .. import _interface

def paralog_pi(align, struct_p, struct_i, max_missing=0.0):
    """
    Compute diversity statistics for a gene family. Based on
    Innan (*Genetics* 2003 **163**:803-810). An
    estimate of genetic diversity is provided for every paralog and for
    every pair of paralogs, provided that enough non-missing data is
    available (at least 2 samples are required). Note that sites with
    more than two alleles are always considered.

    :param align: an :class:`.Align` containing the sequence of all
        available paralog for all samples. The outgroup is ignored.

    :param struct_p: a :class:`.Structure` providing the organisation of
        sequences in paralogs. This structure must have a ploidy of 1 (no
        individual structure). Clusters, if defined, are ignored. The
        sequences of all individuals for a given paralog should be grouped
        together in that structure. There might be a different number of
        sequences per paralog due to missing data.

    :param struct_i: a :class:`.Structure` providing the organisation of
        sequences in individuals. This structure must have a ploidy of 1
        (no individual structure). Clusters, if defined, are ignored. The
        sequences of all paralogs for a given individual should be grouped
        together in that structure. There might be a different number of
        sequences per individual due to missing data.

    :param max_missing: maximum relative proportion of missing data (if there are
        more missing data at a site, the site is ignored altogether).

        .. note::
            Here, *max_missing* is a relative proportion to allow
            using the same value for different alignments that might
            not have the same number of samples (to avoid reevaluating
            *max_missing* if the user wants the same maximum rate
            of missing data). In other functions, *max_missing* is
            the maximum *number* of missing and is an integer.

    :return: A new :class:`.ParalogPi` instance which provides methods
        to access the number of used sites and the diversity for each
        paralog/paralog pair.
    """
    pp = ParalogPi()
    pp.setup(struct_p, struct_i, max_missing)
    pp.process_align(align)
    return pp

class ParalogPi(object):
    """
    Compute diversity statistics for a gene family.
    See :func:`.paralog_pi` for more details. This class
    can be used directly (1) to analyse data with more efficiency (by
    reusing the same instance) or (2) to combine data from different
    alignments, or (3) for pass individual sites. Do first call
    :meth:`.setup`.
    """
    def __init__(self):
        self._obj = _eggwrapper.ParalogPi()
        self._req = 0

    def setup(self, struct_p, struct_i, max_missing=0.0):
        """
        Specify the structure in paralog and individuals. The arguments
        are as as described for :func:`.paralog_pi`.
        Only this method resets the instance.
        """
        if struct_p.ploidy != 1: raise ValueError('ploidy is required to be 1')
        if struct_i.ploidy != 1: raise ValueError('ploidy is required to be 1')
        if max_missing < 0 or max_missing > 1: raise ValueError('max_missing out of bound')
        self._req = struct_p.req_ns
        self._obj.reset(struct_p._obj, struct_i._obj, max_missing)

    def process_align(self, aln):
        """
        Process an alignment. The alignment must match the structure
        passed to :meth:`.setup`.
        Diversity estimates are incremented (no reset).

        :param aln: an :class:`.Align` instance.
        """
        if not isinstance(aln, _interface.Align): raise TypeError('expect an Align instance')
        if aln.ns < 2: return
        if aln.ns > self._req: raise ValueError('unsufficient number of samples in alignment')
        for site in aln.iter_sites():
            self._obj.load(site._obj)

    def process_site(self, site):
        """
        Process a site. The site must match the structure passed to
        :meth:`.setup`. Diversity estimates are incremented (no reset).

        :param site: a :class:`.Site` instance.
        """
        if site.ns < 2: return
        if site.ns > self._req: raise ValueError('unsufficient number of samples in site')
        self._obj.load(site._obj)

    def num_sites(self, *args):
        """
        num_sites([i[, j]])

        Number of sites with data. Number of sites: with any data
        (without arguments), with data for
        paralog *i* (if only *i* specified), with data for the pair
        of paralogs *i* and *j* (if both specified). This counts the number
        of sites which have not been excluded based on the *num_missing* argument,
        and which have at least one pair of samples for the considered sample.
        """
        if len(args) == 0:
            return self._obj.num_sites_tot()
        elif len(args) == 1:
            if args[0] < 0 or args[0] >= self._obj.num_paralogs(): raise IndexError('invalid paralog index')
            return self._obj.num_sites_paralog(args[0])
        elif len(args) == 2:
            if args[0] < 0 or args[0] >= self._obj.num_paralogs(): raise IndexError('invalid paralog index')
            if args[1] < 0 or args[1] >= self._obj.num_paralogs() or args[1] == args[0]: raise IndexError('invalid paralog index')
            return self._obj.num_sites_pair(args[0], args[1])
        else:
            raise ValueError('invalid number of arguments')

    def Piw(self, i):
        """
        Within-paralog diversity for paralog *i*.
        """
        if i<0 or i>=self._obj.num_paralogs(): raise IndexError('invalid paralog index')
        if self._obj.num_sites_paralog(i) < 1: return None
        return self._obj.Piw(i)

    def Pib(self, i, j):
        """
        Between-paralog diversity for paralogs *i* and *j*.
        """
        if i<0 or i>=self._obj.num_paralogs(): raise IndexError('invalid paralog index')
        if j<0 or j>=self._obj.num_paralogs() or j==i: raise IndexError('invalid paralog index')
        if self._obj.num_sites_pair(i, j) < 1: return None
        return self._obj.Pib(i, j)



"""
    Copyright 2008-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import re, sys
from io import StringIO
_staden_table = str.maketrans(' -*', '?N-')
from .. import eggwrapper as _eggwrapper
from .. import _interface, alphabets

special_DNA = alphabets.Alphabet('char', alphabets.DNA.get_alleles()[0], alphabets.DNA.get_alleles()[1], case_insensitive=True, name='DNA with dot')
special_DNA.add_missing('.')
special_DOT = special_DNA._obj.get_code('.')
special_MISSING = special_DNA._obj.get_code('?')
MISSING = alphabets.DNA._obj.get_code('?')

def from_clustal(string, alphabet):
    """
    Import a clustal-formatted alignment. The input format is the one
    generated and used by `ClustalW <http://www.clustal.org/clustal2/>`_.

    :param string: clustal-formatted sequence alignment.
    :param alphabet: an :class:`.Alphabet` instance.
    :return: A new :class:`.Align` instance.
    """

    if not isinstance(alphabet._obj, _eggwrapper.CharAlphabet):
        raise ValueError('invalid alphabet for parsing fasta data: {0}'.format(alphabet.name))
    stream = StringIO(string)

    # the first line must start by CLUSTAL W or CLUSTALW
    line = stream.readline()
    c = 1
    if (line[:8] != 'CLUSTALW' and
        line[:9] != 'CLUSTAL W' and
        line.strip() != 'CLUSTAL'): raise ValueError('invalid clustalw string (line: {0})'.format(c))

    # start reading blocks
    cnt = _eggwrapper.DataHolder(False)
    line = stream.readline()

    # initialize names list
    names = []

    while True:
        
        # skip empty lines
        while line != '' and line.strip() == '':
            line = stream.readline()
            c += 1

        # detect end of file
        if line == '': break

        # read a block
        block_length = None

        # read sequences
        while True:

            # conservation line
            if line[0] == ' ':
                if block_length is None: raise ValueError('invalid clustalw string (line: {0})'.format(c))
                line = stream.readline()
                c += 1
                if line == '': break # end of file
                if line.strip() != '': raise ValueError('invalid clustalw string (line: {0})'.format(c))
                break

            # reads sequence
            bits = line.split()

            # sequence line
            if len(bits) < 2 or len(bits) > 3:
                raise ValueError('invalid clustalw string (line: {0})'.format(c))

            name = bits[0]
            seq = bits[1]
            if block_length is None:
                block_length = len(seq)
            elif block_length != len(seq):
                raise ValueError('invalid clustalw string (line: {0})'.format(c))

            # adds the sequence to the container (new sequence)
            if name not in names:
                pos = len(names)
                cnt.set_nsam(pos + 1)
                cnt.set_name(pos, name)
                cnt.set_nsit_sample(pos, len(seq))
                for i, v in enumerate(seq):
                    cnt.set_sample(pos, i, alphabet._obj.get_code(v))
                names.append(name)

            # adds the sequence (continuing old sequence)
            else:
                pos = names.index(name)
                cur = cnt.get_nsit_sample(pos)
                cnt.set_nsit_sample(pos, cur + len(seq))
                for i, v in enumerate(seq):
                    cnt.set_sample(pos, cur+i, alphabet._obj.get_code(v))

            if len(bits) == 3:
                try:
                    i = int(bits[2])
                except ValueError:
                    raise ValueError('invalid clustalw string (line: {0})'.format(c))
                if cnt.get_nsit_sample(pos) != i:
                    raise ValueError('invalid clustalw string (line: {0})'.format(c))

            # checks next line
            line = stream.readline()
            c += 1
              
            # empty (or absent) conservation line is caught by this line
            if line.strip() == '': break

    if not cnt.is_equal(): raise ValueError('invalid clustalw string (unequal sequences)')
    return _interface.Align._create_from_data_holder(cnt, alphabet)

def from_staden(string, keep_consensus=False):
    """
    Import data from the Staden assembly package.
    Process the output file of the GAP4 program of the
    `Staden <http://staden.sourceforge.net/>`_ package.

    :param string: input string.
    :param keep_consensus: don't delete consensus sequence.
    :return: An :class:`.Align` instance.

    The input string should have been generated from a contig alignment by
    the GAP4 contig editor, using the command "dump contig to file". The
    sequence named ``CONSENSUS``, if present, is automatically removed
    unless the option *keep_consensus* is used.

    Staden's default convention is followed:

    * ``-`` codes for an unknown base and is replaced by ``N``.
    * ``*`` codes for an alignment gap and is replaced by ``-``.
    * ``.`` represents the same sequence than the consensus at that
      position.
    * White space represents missing data and is replaced by ``?``.
    """

    # get shift from the first CONSENSUS line
    mo = re.search('(        CONSENSUS +)[A-Za-z\-\*]', string)
    if mo is None: raise ValueError('invalid staden contig dump file')
    shift = len(mo.group(1))

    # split lines and identify blocks (based on empty lines)
    lines = string.splitlines()
    empty_lines = [i for i, v in enumerate(lines) if v == '']
    empty_lines.insert(0, -1) # emulate a white line immediately before first line
    empty_lines.append(len(lines))
    blocks = [lines[empty_lines[i]+2 : empty_lines[i+1]] for i in range(len(empty_lines) - 1)]
        # +1 to read after each blank line
        # +2 to skip the first line of each block

    # initialize variables
    align = _eggwrapper.DataHolder(False)
    ns = 0
    currpos = 0
    IDs = []

    # process all blocks
    for block in blocks:
        for line in block:
            ID = line[:7].strip()
            name = line[8:shift].strip()
            sequence = line[shift:]
            sequence = sequence.translate(_staden_table)
            block_length = len(sequence)

            if ID not in IDs:
                ns += 1
                align.set_nsam(ns)
                align.set_name(ns-1, name)
                align.set_nsit_sample(ns-1, currpos)
                for i in range(currpos): align.set_sample(ns-1, i, special_MISSING)
                pos = ns - 1
                IDs.append(ID)
            else:
                pos = IDs.index(ID)

            align.set_nsit_sample(pos, currpos + len(sequence))
            for i, v in enumerate(sequence):
                align.set_sample(pos, currpos + i, special_DNA._obj.get_code(v))

        currpos += block_length

    # equalize
    for i in range(ns):
        n = align.get_nsit_sample(i)
        align.set_nsit_sample(i, currpos)
        for j in range(n, currpos): align.set_sample(i, j, special_MISSING)
    align.set_is_matrix(True)

    # undot
    idx = IDs.index('')
    for i in range(ns):
        if i != idx:
            for j in range(currpos):
                if align.get_sample(i, j) == special_DOT:
                    align.set_sample(i, j, align.get_sample(idx, j))

    # remove consensus
    if not keep_consensus:
        align.del_sample(idx)

    # return
    return _interface.Align._create_from_data_holder(align, alphabets.DNA)

def from_genalys(string):
    """
    Import Genalys data.
    Genalys was a proprietary program to process sequencing reads, perform
    assembly and detect polymorphism.
    Convert Genalys-formatted sequence alignment files to fasta. This
    function imports files generated through the option "Save SNPs" of
    Genalys 2.8.

    :param string: input data as a Genalys-formatted string.
    :return: An :class:`.Align` instance.
    """

    stream = StringIO(string)

    insertions = []
    flag = False

    for line in stream:
        line = line.split("\t")

        if len(line) > 1 and line[0] == "Polymorphism":
            flag = True

        if len(line) > 1 and line[0] == "IN" and flag:
            insertions.extend(line[1].split("/"))

    if len(insertions) > 0:
        tp = insertions[0].split("_")
        if len(tp) == 1:
            tp = tp[0].split(".")
            if len(tp) == 1:
                tp.append("1")
        finsertions = [tp]

    for i in insertions:
        i = i.split("_")
        if len(i) == 1:
            tp = tp[0].split(".")
            if len(tp) == 1:
                i.append("1")
        if i[0] != finsertions[-1][0]:
            finsertions.append(i)
        finsertions[-1][1] = i[1]
    
    if len(insertions) > 0:
        insertions = finsertions

    stream.close()
    stream = StringIO(string)

    names = []
    sequences = []
    maxlen = 0

    for line in stream:
        line = line.split("\t")
            
        if len(line) > 1:
            bidon = re.match(".+\.ab1$", line[1])
            if bidon != None:
                names.append(line[1])
                sequences.append("")
                index = 6
                for i in range(10):
                    if line[i] == "F" or line[i] == "R":
                        index = i+1
                        break
                if line[index] != "":
                    beginning = int(line[index]) - 1
                    for i in insertions:
                        if int(i[0]) <= beginning:
                            beginning = beginning + int(i[1])
                        else:
                            break
                    for i in range(beginning):
                        sequences[-1]= sequences[-1] + "?"
                sequences[-1] = sequences[-1] + line[-1].rstrip("\n")
                if len(sequences[-1]) > maxlen:
                    maxlen = len(sequences[-1])

    data = _eggwrapper.DataHolder(True)
    data.set_nsam(len(sequences))
    data.set_nsit_all(maxlen)

    for i in range(len(sequences)):
        data.set_name(i, names[i])
        sequences[i] = sequences[i].replace("_", "-")
        for j, v in enumerate(sequences[i]): data.set_sample(i, j, alphabets.DNA._obj.get_code(v))
        for j in range(len(sequences[i]), maxlen): data.set_sample(i, j, MISSING)

    return _interface.Align._create_from_data_holder(data, alphabets.DNA)

def get_fgenesh(string, locus='locus'):
    """
    Import fgenesh data.

    :param fname: a string containing fgenesh ouput.
    :parma locus: locus name.
    :return: A list of gene and CDS features
        represented by dictionaries. Note that 5' partial features
        might not be in the appropriate frame and that it can be
        necessary to add a ``codon_start`` qualifier.
    """

    # supports for mac/windows files
    string = '\n'.join(string.splitlines())

    # gets the feature table
    try:
        data_sub = string.split('   G Str   Feature   Start        End    Score           ORF           Len\n')[1].split('Predicted protein(s):\n')[0]
    except IndexError:
        raise ValueError('invalid fgenesh format')
    data_sub = data_sub.split('\n\n')

    # edit
    del data_sub[-1]
    data_sub[0]= '\n'.join(data_sub[0].split('\n')[1:])

    # iteratively grabs the features
    features = {}
    for i in data_sub:
        pos = []
        start = 1
        rank = '---'
        strand = '---'
        for j in i.split('\n'):
            a = re.search(' ?[0-9]+ ([+|-])      (TSS|PolA) +([0-9]+)', j)
            b = re.search(' ?([0-9]+) ([+|-]) + ([0-9])+ CDS(o|f|i|l) +([0-9]+) - +([0-9]+) +[-\.0-9]+ + ([0-9]+)', j)
            if b:
                if b.group(3) == "1":
                    if int(b.group(5)) == int(b.group(7)): start= 1
                    elif int(b.group(5)) == (int(b.group(7))-1): start= 2
                    elif int(b.group(5)) == (int(b.group(7))-2): start= 3
                    else: raise ValueError('invalid fgenesh format')
                pos.append( [int(b.group(5))-1, int(b.group(6))-1 ] )
                rank = b.group(1)
                if b.group(2) == '+': strand = 'plus'
                else: strand = 'minus'

        features['cds'+rank] ={
            'gene': locus+'_'+rank,
            'strand': strand,
            'pos': pos,
            'type': 'CDS',
            'note': 'fgenesh prediction'
        }

        features['gene'+rank] ={
            'gene': locus+'_'+rank,
            'strand': strand,
            'pos': [[ pos[0][0], pos[-1][1] ]],
            'type': 'gene',
            'note': 'fgenesh prediction'
        }
        
    # gets the sequence section
    try:
        data_sub = string.split('   G Str   Feature   Start        End    Score           ORF           Len\n')[1].split('Predicted protein(s):\n')[1].split('>')
    except IndexError:
        raise ValueError('invalid fgenesh format')
    del data_sub[0]

    if ( (2*len(data_sub) != len(features)) and
           (len(data_sub) != len(features)) ) : raise ValueError('invalid fgenesh format')

    # returns the sequences as a table
    return list([features[i] for i in features])



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest, egglib, pathlib
path = pathlib.Path(__file__).parent / '..' / 'data'

class GFF3_test(unittest.TestCase):

    def setUp(self):
        fname='Populus_trichocarpa_short_thinned.gff3'
        self.gff3=egglib.io.GFF3(path / fname)

    def tearDown(self):
        del self.gff3

    def test_GFF3_T(self):
        self.assertIsInstance(self.gff3, egglib.io._gff3.GFF3)

    def test_num_top_features_T(self):
        ntf=self.gff3.num_top_features
        self.assertEqual(ntf, 16212)

    def test_num_tot_features_T(self):
        nttf=self.gff3.num_tot_features
        self.assertEqual(nttf, 22992)

    def test_feature_iter_GFF3_T(self):
        fname='Osativa_193_gene_thinned.gff3'
        gff3=egglib.io.GFF3(path / fname)
        f_iter=gff3.iter_features('Chr1', 1080000, 2510000)
        for feat in f_iter: 
            self.assertIsInstance(feat, egglib.io.Gff3Feature)
            self.assertEqual(feat.ID, 'LOC_Os01g02940')
            self.assertEqual(feat.attributes['Name'], 'LOC_Os01g02940')
            self.assertEqual(feat.source, 'phytozome8_0')
            break

    def test_mini_GFF3(self):
        A = egglib.io.GFF3.from_string(open(path / 'a.gff3').read())
        B = egglib.io.GFF3(path / 'a.gff3')
        C = egglib.io.GFF3(path / 'b.gff3')

        self.assertEqual(A.version, '3')
        self.assertEqual(A.regions, {'1': (1-1, 500000-1)})
        self.assertEqual(A.feature_ontology, ['X001', 'X002'])
        self.assertEqual(A.attribute_ontology, ['X003'])
        self.assertEqual(A.source_ontology, ['X004', 'X005'])
        self.assertEqual(A.species, 'A species')
        self.assertEqual(A.genome_build, ('source', 'buildName'))
        self.assertEqual(A.sequences.fasta(), '>1\nAAAGCGCCGGGCGCGCCC\n')
        self.assertEqual(B.sequences.fasta(), '>1\nAAAGCGCCGGGCGCGCCC\n')
        self.assertEqual(C.sequences.fasta(), '>1\nAAAGCGCCGGGCGCGCCC\n>2\nTCCGCGCG\n')
        self.assertEqual(A.num_seqid, 2)
        self.assertEqual(A.list_seqid, ['1', 'chr3'])
        self.assertEqual(A.num_top_features, 6)
        self.assertEqual(A.num_tot_features, 24)

        it = [(i.seqid, i.start, i.end) for i in A]
        self.assertEqual(it, [('1', 295054, 303569), ('1', 494999, 549999), ('chr3', 0, 22), ('chr3', 27, 34), ('chr3', 56, 74), ('chr3', 87, 101)])

        it = [(i.seqid, i.start, i.end) for i in A.iter_features()]
        self.assertEqual(it, [('1', 295054, 303569), ('1', 494999, 549999), ('chr3', 0, 22), ('chr3', 27, 34), ('chr3', 56, 74), ('chr3', 87, 101)])

        it = [(i.seqid, i.start, i.end) for i in A.iter_features(seqid='1')]
        self.assertEqual(it, [('1', 295054, 303569), ('1', 494999, 549999)])

        it = [(i.seqid, i.start, i.end) for i in A.iter_features(seqid='1', start=295200, end=301000, all_features=True)]
        self.assertEqual(it, [('1', 295251, 295332), ('1', 295251, 295737), ('1', 297367, 297726), ('1', 298780, 298876), ('1', 299251, 299681), ('1', 300452, 300607)])

        it = [(i.seqid, i.start, i.end) for i in A.iter_features(seqid='chr3', start=25, end=100)]
        self.assertEqual(it, [('chr3', 27, 34), ('chr3', 56, 74)])

        [feat] = list(A.iter_features(seqid='1', start=298780, end=298876, all_features=True))

        self.assertEqual(feat.seqid, '1')
        self.assertEqual(feat.source, 'jgi')
        self.assertEqual(feat.type, 'exon')
        self.assertEqual(feat.start, 298780)
        self.assertEqual(feat.end, 298876)
        self.assertEqual(feat.score, None)
        self.assertEqual(feat.strand, '+')
        self.assertEqual(len(feat.segments), 1)
        self.assertEqual(feat.segments[0], (298780, 298876, None))
        self.assertEqual(len(feat.all_parts), 1)
        self.assertEqual(len(feat.descendants), 0)
        self.assertEqual(len(feat.parents), 1)
        self.assertEqual(feat.parents[0].ID, 'transcript:POPTR_0001s00390.2')
        dest = []
        feat._ultimate_parents(dest)
        self.assertEqual(len(dest), 1)
        self.assertEqual(dest[0].ID, 'gene:POPTR_0001s00390')
        ref = {
             'Parent': ('transcript:POPTR_0001s00390.2',),
             'Name': 'POPTR_0001s00390.2.exon4',
             'constitutive': '1',
             'ensembl_end_phase': '1',
             'ensembl_phase': '0',
             'exon_id': 'POPTR_0001s00390.2.exon4',
             'rank': '4',
             'version': '1',
             'ID': None,
             'Alias': None,
             'Target': None,
             'Gap': None,
             'Derives_from': None,
             'Note': None,
             'Dbxref': None,
             'Ontology_term': None,
             'Is_circular': False}
        self.assertEqual(feat.attributes, ref)

    def test_CDS_feat(self):
        gff3 = egglib.io.GFF3.from_string(open(path / 'a.gff3').read())

        for feat in gff3.iter_features(all_features=True):
            feat.seqid, type(feat.start), feat.end, feat.type

        feats = list(gff3.iter_features(seqid='1', start=295333, end=301972, all_features=True))
        cds =  []
        for feat in feats:
            if feat.type == 'CDS': cds.append(feat)
        self.assertEqual(len(cds), 1)
        cds = cds[0]
        for a, b, c in cds.segments: self.assertIsInstance(c, int)

    def test_feat_count(self):
        fname = path / 'Populus_trichocarpa_short_thinned.gff3'
        gff = egglib.io.GFF3(fname)
        types = {}
        for seqid in '12':
            for feat in gff.iter_features(seqid, 0, 100000000000000):
                if feat.type not in types: types[feat.type] = 0
                types[feat.type] += 1
        ctrl = {}
        f = open(fname)
        for line in f:
            if line[0] != '#':
                bits = line.split('\t')
                assert len(bits) == 9
                tp = bits[2]
                attrs = bits[8]
                P = None
                for attr in attrs.split(';'):
                    k, v = attr.split('=')
                    if k == 'Parent':
                        assert P is None
                        P = v
                if P is None:
                    if tp not in ctrl: ctrl[tp] = 0
                    ctrl[tp] += 1
        self.assertDictEqual(types, ctrl)

    def test_phase(self):
        # Source of dataset:
        # https://www.gencodegenes.org/human/
        # https://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_human/release_43/gencode.v43.annotation.gff3.gz
        # (only beginning of file)

        gff = egglib.io.GFF3(str(path / 'gencode.v43.annotation_head.gff3'))
        for feat in gff.iter_features(all_features=True):
            if feat.type in ['CDS', 'start_codon', 'stop_codon']:
                for (start, stop, phase) in feat.segments:
                    self.assertIsNotNone(phase)
            else:
                for (start, stop, phase) in feat.segments:
                    self.assertIsNone(phase)

class GFF3Feature_test(unittest.TestCase):
    def setUp(self):
        fname='Osativa_193_gene_thinned.gff3'
        self.gff3=egglib.io.GFF3(path / fname)
        self.f_iter=self.gff3.iter_features('Chr1', 1080000, 2510000)

    def tearDown(self):
        del self.gff3

    def test_GFF3Feature_T(self):
        for feat in self.f_iter: 
            self.assertIsInstance(feat, egglib.io.Gff3Feature)
        
    def test_GFF3Feature_property_T(self):
        for feat in self.f_iter: 
            self.assertEqual(feat.attributes, {'ID': 'LOC_Os01g02940', 'Name': 'LOC_Os01g02940',
                'Alias': None, 'Parent': None, 'Target': None,
                'Gap': None, 'Derives_from': None, 'Note': None,
                'Dbxref': None, 'Ontology_term': None,
                'Is_circular': False})
            self.assertEqual(feat.start, 1089742)
            self.assertEqual(feat.end, 1093527)
            self.assertEqual(feat.seqid, 'Chr1')
            self.assertEqual(feat.type, 'gene')
            self.assertEqual(feat.score, None)
            self.assertEqual(feat.strand, '-')
            self.assertEqual(feat.ID, 'LOC_Os01g02940')
            self.assertEqual(feat.source, 'phytozome8_0')
            self.assertEqual(len(feat.segments), 1)
            self.assertEqual(feat.segments, [(1089742, 1093527, None)])
            self.assertEqual(feat.parents, [])
            self.assertEqual(len(feat.descendants), 5)
            self.assertEqual(len(feat.all_parts), 49)
            break

    def test_get_part_T(self):
        for feat in self.f_iter: 
            self.assertIsInstance(feat.descendants[0], egglib.io.Gff3Feature)

    def test_get_part_E(self):
        for feat in self.f_iter: 
            with self.assertRaises(IndexError):
                feat.descendants[10]

    def test_get_parent_E(self):
        for feat in self.f_iter: 
            with self.assertRaises(IndexError):
                feat.parents[10]



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import egglib, unittest, pathlib
path = pathlib.Path(__file__).parent / '..' / 'data'

class Freq_test(unittest.TestCase):
    def test_Freq_T(self):
        freq=egglib.Freq()
        self.assertEqual(str(type(freq)), "<class 'egglib._freq.Freq'>")
        
    def test_process_site_T(self):
        freq=egglib.Freq()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        site=egglib.Site()
        site.from_align(aln, 2213)
        na_b= freq.num_alleles
        freq.from_site(site)
        na_a= freq.num_alleles
        self.assertTrue(na_a>na_b)

    def test_from_site_E(self):
        freq=egglib.Freq()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        sub = aln.subset(range(10,20))
        struct=egglib.struct_from_labels(aln)
        site=egglib.Site()
        site.from_align(sub, 20)

        with self.assertRaises(ValueError):
            freq.from_site(site, struct=struct)

        site.from_align(aln, 2213)
        freq.from_site(site, struct=struct)

        a, b = struct.as_dict()
        a[None][None]['ERROR'] = (100,)
        struct = egglib.struct_from_dict(a, b)
        with self.assertRaises(ValueError):
            freq.from_site(site, struct=struct)

    def test_from_list_T(self):
        freq=egglib.Freq()
        self.assertEqual(freq.num_alleles, 0)
        my_list=[[[84, 16, 32],[ 7, 28,  0],[ 0,  0, 14],[12, 64,  5]]]
        freq.from_list(my_list, None)
        self.assertEqual(freq.num_alleles, 3)
        self.assertEqual(str(type(freq)), "<class 'egglib._freq.Freq'>")

    def test_from_list_E(self):
        freq=egglib.Freq()
        my_list=[[[84, 16, 32],[ 7, 28,  0],[ 0,  0, 14]]]
        freq.from_list(my_list, None)
        with self.assertRaises(ValueError):
            freq.from_list(my_list, [])
        with self.assertRaises(ValueError):
            my_list=[[[2, 5, 3], [1, 1, 8,5,6], [2, 5, 1]],[[11, 1, 0,10], [4, 0, 0]]]
            freq.from_list(my_list, None)
        with self.assertRaises(ValueError):
            my_list=[]
            freq.from_list(my_list, None)
        with self.assertRaises(ValueError):
            my_list=[[[84, 16, 32],[ 7, 28,  0],[ 0,  0, 14]]]
            freq.from_list(my_list, None, geno_list=[(0, 0), (0, 2), (2, 0), (2, 2)])
        with self.assertRaises(ValueError):
            my_list=[[[2, 5, 3], [1, 1, 8], [2, 5, 1]],[[11, 1, 0], [4, 0, 0]]]
            geno_lst=[(0,0), (0,2,0), (2,0,2)]
            freq.from_list(my_list, None, geno_list=geno_lst)
        with self.assertRaises(ValueError):
            my_list=[[[2, 5, 3], [1, 1, 8], [2, 5, 1]],[[11, 1, 0], [4, 0, 0]]]
            geno_lst=[(), (), ()]
            freq.from_list(my_list, None, geno_list=geno_lst)
        with self.assertRaises(ValueError):
            geno_lst=[(0,2), (0,2), (0,2)]
            freq.from_list(my_list, None, geno_list=geno_lst)

    def test_from_vcf_T(self):
        freq=egglib.Freq()
        vcf = egglib.io.VcfParser(str(path / 'example1.vcf'))
        na_b= freq.num_alleles
        for i, v in enumerate(vcf):
            freq.from_vcf(vcf)
            na_a= freq.num_alleles
        self.assertEqual(str(type(freq)), "<class 'egglib._freq.Freq'>")
        self.assertTrue(na_a>na_b)

    def test_from_vcf_E(self):
        freq=egglib.Freq()
        vcf = egglib.io.VcfParser(str(path / 'exemple1_F.vcf'))
        na_b= freq.num_alleles
        with self.assertRaises(ValueError):
            for i, v in enumerate(vcf):
                freq.from_vcf(vcf)

    def test_ploidy_T(self):
        freq=egglib.Freq()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        site=egglib.Site()
        site.from_align(aln, 2213)
        freq.from_site(site)
        self.assertEqual(freq.ploidy,1)

    def test_num_alleles_T(self):
        freq=egglib.Freq()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        site=egglib.Site()
        site.from_align(aln, 2213)
        freq.from_site(site)
        self.assertEqual(freq.num_alleles,2)

    def test_num_genotypes_T(self):
        freq=egglib.Freq()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        site=egglib.Site()
        site.from_align(aln, 2213)
        freq.from_site(site)
        self.assertEqual(freq.num_genotypes, 2)

    def test_num_clusters_T(self):
        freq=egglib.Freq()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        site=egglib.Site()
        site.from_align(aln, 2213)
        freq.from_site(site)
        self.assertEqual(freq.num_clusters, 1)
    
    def test_num_populations_T(self):
        freq=egglib.Freq()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        site=egglib.Site()
        site.from_align(aln, 2213)
        freq.from_site(site)
        self.assertEqual(freq.num_populations, 1)
        
    def test_genotype_T(self):
        my_list=[[[2, 5, 3], [1, 1, 8], [2, 5, 1]],[[11, 1, 0], [4, 0, 0]]]
        freq = egglib.freq_from_list(my_list, [7, 4, 1], geno_list=[(0, 0), (0, 2), (2, 2)], alphabet=egglib.alphabets.positive_infinite)
        self.assertEqual(freq.genotype(2), (2,2))

    def test_genotype_E(self):
        my_list=[[[2, 5, 3], [1, 1, 8], [2, 5, 1]],[[11, 1, 0], [4, 0, 0]]]
        freq = egglib.freq_from_list(my_list, [7, 4, 1],geno_list=[(0, 0), (0, 2), (2, 2)], alphabet=egglib.alphabets.positive_infinite)
        with self.assertRaises(IndexError):
            freq.genotype(100)

    def test_freq_allele_T(self):
        my_list=[[[2, 5, 3], [1, 1, 8], [2, 5, 1]],[[11, 1, 0], [4, 0, 0]]]
        freq = egglib.freq_from_list(my_list, [7, 4, 1],geno_list=[(0, 0), (0, 2), (2, 2)], alphabet=egglib.alphabets.positive_infinite)
        self.assertEqual(freq.freq_allele(0),52)
        
    def test_freq_allele_E(self):
        my_list=[[[2, 5, 3], [1, 1, 8], [2, 5, 1]],[[11, 1, 0], [4, 0, 0]]]
        freq = egglib.freq_from_list(my_list, [7, 4, 1],geno_list=[(0, 0), (0, 2), (2, 2)], alphabet=egglib.alphabets.positive_infinite)
        with self.assertRaises(IndexError):
            freq.freq_allele(100)
        
    def test_freq_genotype_T(self):
        my_list=[[[2, 5, 3], [1, 1, 8], [2, 5, 1]],[[11, 1, 0], [4, 0, 0]]]
        freq = egglib.freq_from_list(my_list, [7, 4, 1],geno_list=[(0, 0), (0, 2), (2, 2)], alphabet=egglib.alphabets.positive_infinite)
        self.assertEqual(freq.freq_genotype(0),20)

    def test_freq_genotype_E(self):
        my_list=[[[2, 5, 3], [1, 1, 8], [2, 5, 1]],[[11, 1, 0], [4, 0, 0]]]
        freq = egglib.freq_from_list(my_list, [7, 4, 1],geno_list=[(0, 0), (0, 2), (2, 2)], alphabet=egglib.alphabets.positive_infinite)
        with self.assertRaises(IndexError):
            freq.freq_genotype(100)

    def test_eff_samples_T(self):
        freq = egglib.Freq()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        site = egglib.Site()
        site.from_align(aln, 2213)
        freq.from_site(site, struct=egglib.struct_from_labels(aln))
        self.assertEqual(freq.nseff(), 44)
        self.assertEqual(freq.nieff(), 44)
        self.assertEqual(freq.ploidy, 1)
        d = {}
        for i in range(28):
            d[str(i)] = [i*2, i*2+1]
        struct = egglib.struct_from_dict({None: {None: d}}, {})
        freq.from_site(site, struct=struct)
        self.assertEqual(freq.nseff(), 43)
        self.assertEqual(freq.nieff(), 16)
        self.assertEqual(freq.ploidy, 2)

    def test_freq(self):
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        site = egglib.site_from_align(aln, 360)
        frq = egglib.freq_from_site(site)
        alleles = []
        array = site.as_list()
        for i in array:
            if i not in alleles: alleles.append(i)

        self.assertEqual(frq.num_alleles, len(alleles))
        self.assertEqual(frq.num_genotypes, len(alleles))
        self.assertEqual(frq.ploidy, 1)
        self.assertEqual([frq.freq_allele(i) for i in range(frq.num_alleles)], [array.count(i) for i in alleles])
        self.assertEqual([frq.freq_genotype(i) for i in range(frq.num_genotypes)], [array.count(i) for i in alleles])
        self.assertEqual([frq.allele(i) for i in range(frq.num_alleles)], alleles)
        self.assertEqual([frq.genotype(i) for i in range(frq.num_genotypes)], list(map(tuple, alleles)))

        frq = egglib.freq_from_list(
            [24, 12, 6],
            [1, 0, 0],
            geno_list = [('A', 'A'), ('A', 'C'), ('C', 'C')],
            alphabet = egglib.alphabets.DNA)

        try:
            frq = egglib.freq_from_list(
                [[[14, 6], [3, 46]]],
                [0, 0], alphabet=egglib.alphabets.Alphabet('int', [0], []))
        except ValueError:
            pass
        else:
            raise AssertionError

        frq = egglib.freq_from_list(
            [[[24, 12, 6], [5, 15, 7]]],
            [1, 0, 0],
            geno_list = [('A', 'A'), ('A', 'C'), ('C', 'C')],
            alphabet = egglib.alphabets.DNA)

        self.assertEqual(frq.num_alleles, 2)
        self.assertEqual(frq.num_genotypes, 3)
        self.assertEqual(frq.ploidy, 2)
        self.assertEqual([frq.freq_allele(i) for i in range(frq.num_alleles)], [(24+5)*2 + (12+15), (12+15) + (6+7)*2])
        self.assertEqual([frq.freq_genotype(i) for i in range(frq.num_genotypes)], [24+5, 12+15, 6+7])
        self.assertEqual([frq.allele(i) for i in range(frq.num_alleles)], ['A', 'C'])
        self.assertEqual([frq.genotype(i) for i in range(frq.num_genotypes)], [('A', 'A'), ('A', 'C'), ('C', 'C')])
